# adapted from https://stackoverflow.com/questions/42697933/colormap-with-maximum-distinguishable-colours
import math

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap


def _generate_colormap_(number_of_distinct_colors: int,
                        cmap: str = "hsv"):
    number_of_shades = 7
    number_of_distinct_colors_with_multiply_of_shades = int(
        math.ceil(number_of_distinct_colors / number_of_shades) * number_of_shades)

    # Create an array with uniformly drawn floats taken from <0, 1) partition
    linearly_distributed_nums = np.arange(
        number_of_distinct_colors_with_multiply_of_shades) / number_of_distinct_colors_with_multiply_of_shades

    # We are going to reorganise monotonically growing numbers in such way that there will be single array with
    # saw-like pattern but each saw tooth is slightly higher than the one before First divide
    # linearly_distributed_nums into number_of_shades sub-arrays containing linearly distributed numbers
    arr_by_shade_rows = linearly_distributed_nums.reshape(number_of_shades,
                                                          number_of_distinct_colors_with_multiply_of_shades // number_of_shades)

    # Transpose the above matrix (columns become rows) - as a result each row contains saw tooth with values slightly
    # higher than row above
    arr_by_shade_columns = arr_by_shade_rows.T

    # Keep number of saw teeth for later
    number_of_partitions = arr_by_shade_columns.shape[0]

    # Flatten the above matrix - join each row into single array
    nums_distributed_like_rising_saw = arr_by_shade_columns.reshape(-1)

    # HSV colour map is cyclic (https://matplotlib.org/tutorials/colors/colormaps.html#cyclic), we'll use this property
    initial_cm = plt.get_cmap(cmap)(nums_distributed_like_rising_saw)

    lower_partitions_half = number_of_partitions // 2
    upper_partitions_half = number_of_partitions - lower_partitions_half

    # Modify lower half in such way that colours towards beginning of partition are darker
    # First colours are affected more, colours closer to the middle are affected less
    lower_half = lower_partitions_half * number_of_shades
    for i in range(3):
        initial_cm[0:lower_half, i] *= np.arange(0.2, 1, 0.8 / lower_half)

    # Modify second half in such way that colours towards end of partition are less intense and brighter
    # Colours closer to the middle are affected less, colours closer to the end are affected more
    for i in range(3):
        for j in range(upper_partitions_half):
            modifier = np.ones(number_of_shades) - initial_cm[lower_half + j * number_of_shades: lower_half + (
                        j + 1) * number_of_shades, i]
            modifier = j * modifier / upper_partitions_half
            initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i] += modifier

    return ListedColormap(initial_cm)



# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, Tuple


class ParsingError(Exception):
    pass

class MolecularSpeciesError(Exception):
    pass

class ReferenceLipidError(Exception):
    pass

class MappingError(Exception):
    pass


class InputDataError(Exception):
    def __init__(self, message):
        super(InputDataError, self).__init__(message)
        self.message = message


class NotComputedError(Exception):
    def __init__(self, attribute_name: str,
                 function_name: Union[str, Tuple[str]] = None,
                 for_subset: bool = False):
        if function_name is None:
            super(NotComputedError, self).__init__(
                attribute_name
            )
        else:
            if for_subset:
                message = f"{attribute_name} has not been computed for {function_name}"
            else:
                message = f"{attribute_name} has not been computed yet,"\
                          f" please call {function_name} first!"
            super(NotComputedError, self).__init__(message)


class CorrelationError(Exception):
    def __init__(self, message, lipids):
        super(CorrelationError, self).__init__(message)

        self.lipids = lipids
        self.message = message


class PartialCorrelationError(Exception):
    def __init__(self, message, solver):
        super(PartialCorrelationError, self).__init__(message)

        self.solver = solver
        self.message = message


class SignificanceTestError(Exception):
    def __init__(self, message):
        super(SignificanceTestError, self).__init__(message)



import importlib.resources as pkg_resources
import pandas as pd
from . import data

with pkg_resources.path(data, "Reactome_ConnectionOfReactionsToMolecule_all.csv") as path:
    REACTOME_REACTION_TO_MOLECULE = pd.read_csv(path)

with pkg_resources.path(data, "Reactome_Others_all_unique.csv") as path:
    REACTOME_OTHERS_UNIQUE = pd.read_csv(path)

with pkg_resources.path(data, "Reactome_ReactionDetailsEnzyme_all.csv") as path:
    REACTOME_REACTION_DETAILS = pd.read_csv(path)

with pkg_resources.path(data, "Reactome_reactions_for_curation.csv") as path:
    REACTOME_REACTION_CURATION = pd.read_csv(path)

with pkg_resources.path(data, "Rhea_ConnectionOfReactionsToMolecule_all.csv") as path:
    RHEA_REACTION_TO_MOLECULE = pd.read_csv(path)

with pkg_resources.path(data, "Rhea_Others_all_unique.csv") as path:
    RHEA_OTHERS_UNIQUE = pd.read_csv(path)

with pkg_resources.path(data, "Rhea_reactions_for_curation.csv") as path:
    RHEA_REACTION_CURATION = pd.read_csv(path)

with pkg_resources.path(data, "Rhea_GeneName_mapping.csv") as path:
    RHEA_MAPPING = pd.read_csv(path).set_index("RHEA_ID")

with pkg_resources.path(data, "standard_lipid_classes.csv") as path:
    STANDARD_LIPID_CLASSES = pd.read_csv(path)



# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Union
import numpy as np
import warnings

# From package
from .lipid import Lipid
from .parser import lipid_parser
from .GenerateLipidNetwork import lipid_attributes


def _get_weighted_lipid_dict(self) -> Dict[str, List[Lipid]]:
    lipid_dict = self.gln.get_lipid_dict()
    for cls in lipid_dict.keys():
        tmp_weights = {}
        # Count how often a lipid from the same dataname exists in the data
        for lip in lipid_dict[cls]:
            if lip.get_dataname() in tmp_weights.keys():
                tmp_weights[lip.get_dataname()] += 1
            else:
                tmp_weights[lip.get_dataname()] = 1
        # Use these counts as factors
        for lipi in range(len(lipid_dict[cls])):
            lipid_dict[cls][lipi].set_div_factor(tmp_weights[lipid_dict[cls][lipi].get_dataname()])

    return lipid_dict


def add_lipid(self, new_lipid: Union[str, List[str], Lipid, List[Lipid]],
              compute_new_edgelist: bool = True,
              is_lm_compatible: bool = False):
    """
    Adding one or multiple new lipid(s) to the network.
    Please note: compute_*_network has to be called separately after
    new lipids have been added

    Parameters
    ----------
    :param new_lipid : Union[str, Iterable[str], Lipid, Iterable[Lipid]]
        Lipid(s) to add to the network. These can be lipid class objects or
        lipid names as str, which can then be parsed into Lipid objects

    :param compute_new_edgelist : bool, optional, default True
        Whether to recompute the edgelist based on the new
        lipid set

    :param prl : bool, optional, default True
        Whether to compute the edgelist in a parallelized fashion

    :param is_lm_compatible: cool, optional, default = False
        When the lipids in the data file are not compatible with the LIPID MAPS nomenclature,
        they can be converted with LipidLynxX. If the parameter is False,
        lipids will be converted before (This can take up to several minutes).
        If True, the lipids won't be converted.
        For more information visit the LipidLynxX website: https://www.lipidmaps.org/lipidlynxx/

    Raises
    ------
    ValueError
        If new_lipid is not a Lipid object or a list of Lipid object
    """
    # Converted lipids
    nl = []

    if isinstance(new_lipid, List):
        if isinstance(new_lipid[0], Lipid):
            for lipid in new_lipid:
                self.gln.add_new_lipid(lipid)
            nl = new_lipid
        if isinstance(new_lipid[0], str):
            for lipid in new_lipid:
                tmp_lipid = lipid_parser(lipid,
                                         self._ref_dict,
                                         is_ll_compatible=is_lm_compatible,
                                         org_name=lipid)
                self.gln.add_new_lipid(tmp_lipid)
                nl.append(tmp_lipid)

    elif isinstance(new_lipid, str):
        # Parse lipid from str
        tmp_lipid = lipid_parser(new_lipid,
                                 self._ref_dict,
                                 is_ll_compatible=is_lm_compatible,
                                 org_name=new_lipid)
        self.gln.add_new_lipid(tmp_lipid)
        nl = [tmp_lipid]
    else:
        if not isinstance(new_lipid, Lipid):
            raise ValueError(
                "'new_lipid' must be an object of class 'Lipid' or 'str'"
                f"or a list of 'Lipid' or 'str' objects, not {type(new_lipid).__name__}"
            )
        self.gln.add_new_lipid(new_lipid)
        nl = [new_lipid]
    # TODO: add computed properties and colours for new_lipid
    for lip in nl:
        lipid_attrs = lipid_attributes(lip)
        for attr in self._attributes_to_plot["nodes"]:
            if self._attr_group_size_[attr] == 0:
                if self.lipid_attributes.get(attr) is not None:
                    if self.last_network_call.get('network', '') == "native":
                        self.lipid_attributes[attr][lip.get_native_string()] = lipid_attrs.get(attr, np.nan)
                    else:
                        self.lipid_attributes[attr][str(lip)] = lipid_attrs.get(attr, np.nan)
    # recompute edgelist:
    if compute_new_edgelist:
        # NOTE: currently edge attributes are NOT reset,
        # because self.network is not necessarily changed
        # self._reset_edges_()
        self.gln.compute_edgelist()
        if self.last_network_call:
            self.last_network_call['force_overwrite'] = True
            self._compute_network_(**self.last_network_call)
        else:
            warnings.warn(
                'No network computed yet!'
            )


def all_molecular_species_lipids(self) -> List[str]:
    """
    Converts the internal "__lipids" dict into a list of (molecular species) string representations of all lipids.
    :return: List of string representations of molecular species lipids
    """
    out_l = []
    lip_dict = self.gln.get_lipid_dict()
    for cls in lip_dict:
        # FIXME: how to get the correct lipid name(s) here?
        out_l += [str(x) for x in lip_dict[cls]]
    return out_l


def all_native_lipids(self) -> List[str]:
    """
    Converts the internal "__lipids" dict into a list of (sum species) string representations of all lipids.
    :return: List of string representations of sum species lipids
    """
    out_l = []
    lip_dict = self.gln.get_lipid_dict()
    for cls in lip_dict:
        out_l += [x.get_native_string() for x in lip_dict[cls]]
    return out_l


def all_lipid_objects(self) -> List[Lipid]:
    out_l = []
    lip_dict = self.gln.get_lipid_dict()
    for cls in lip_dict:
        out_l += [x for x in lip_dict[cls]]
    return out_l


def get_incompatible_lipids(self) -> List[str]:
    """
    Get all incompatible lipids.

    These lipids were not parsed/identified correctly and therefore removed from the data.

    :return: List of removed lipids (column names) from the data.
    """
    return self._incompatible_lipids



import itertools

import numpy as np
import pandas as pd
from typing import Union, Tuple, Dict, List, Optional, Any
from sklearn.linear_model import LinearRegression
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from matplotlib.colors import to_hex
import seaborn as sns
from .lipid import Lipid

N_COLS = 3
FUNCTION_NAMES = {
    Lipid.get_lipid_class: 'LipidClass',
    Lipid.sum_length: 'SumLength',
    Lipid.sum_dbs: 'SumDBs',
    Lipid.sum_ohs: 'SumOHs',
    Lipid.get_backbone: 'Backbone',
    Lipid.get_headgroup: 'Headgroup',
    Lipid.get_category: 'Category',
    Lipid.get_fa_long_chain_base: 'LongChainBase',
    Lipid.get_fa_lengths: 'FALength',
    Lipid.get_fa_dbs: 'FADBs'
}
DEFAULT_COLOUR = to_hex(plt.get_cmap('tab10')(0))


def _get_colours_(n: int) -> List[str]:
    if n < 11:
        colours = plt.get_cmap('tab10').colors
    elif n < 21:
        colours = plt.get_cmap('tab20').colors
    else:
        colours = plt.cm.rainbow(np.linspace(0, 1, n))
    return [to_hex(col) for col in colours]


def _ceiling_division_(a, b):
    return -(a // -b)


def _add_feature_combinations_(feature_matrix, lipid, features, n_combinations):
    for x in itertools.combinations(features, n_combinations):
        combined_feature = ' -- '.join(sorted(x))
        if combined_feature not in feature_matrix.index.values:
            feature_matrix = feature_matrix.append(pd.Series([0] * feature_matrix.shape[1],
                                                             index=feature_matrix.columns,
                                                             name=combined_feature))
        feature_matrix.loc[combined_feature, lipid.get_dataname()] += (1 / lipid.get_div_factor())
    return feature_matrix


def analysis_lipid_substructure(
    self, functions: list, feature_combinations: int = 1,
    combined: bool = False, single_n_combination: bool = False
) -> pd.DataFrame:
    """
    TODO: Tim

    Parameters
    ----------
    :param functions
    :param feature_combinations
    :param combined: bool, default=False
        Combine newly generated lipid substructures with the original lipid data and
        output the combined data frame.
    :param single_n_combination

    Returns
    -------
    :return: pd.DataFrame
        Data frame with lipid substructures as amounts for each sample.
        Optionally, the data frame is combined with the original lipid data.
    """
    # Set lipid weights
    lipid_dict = self._get_weighted_lipid_dict()

    # Construction of the feature weight matrix
    # This will later be multiplied with data matrix
    feature_matrix = self.data.T.drop(self.data.T.index.values)

    for cls in lipid_dict.keys():
        for lip in lipid_dict[cls]:
            # make list of all considered features
            tmp_features = list()
            # *category
            for ele in functions:
                if ele == Lipid.get_fas:
                    # **single
                    for i in ele(lip):
                        tmp_features.append(f"FA: {str(i)}")
                    # **combined
                    if len(ele(lip)) >= 2:
                        for x in itertools.combinations(ele(lip), 2):
                            tmp_features.append(f"FA: {str(sorted(x))}")
                    if len(ele(lip)) >= 3:
                        for x in itertools.combinations(ele(lip), 3):
                            tmp_features.append(f"FA: {str(sorted(x))}")
                    if len(ele(lip)) >= 4:
                        print("Combinations of 4 FA's are not implemented")
                else:
                    tmp_features.append(
                        f"{FUNCTION_NAMES.get(ele, str(ele).split(' ')[1].split('.')[1])} = {str(ele(lip))}"
                    )
            if single_n_combination:
                feature_matrix = _add_feature_combinations_(
                    feature_matrix, lip, tmp_features,
                    feature_combinations
                )
            else:
                count = 1
                while count <= feature_combinations:
                    feature_matrix = _add_feature_combinations_(
                        feature_matrix, lip, tmp_features,
                        count
                    )
                    count += 1
    if combined:
        return feature_matrix.dot(self.data).append(self.data)
    else:
        return feature_matrix.dot(self.data)


def _flatten_(data: pd.DataFrame):
    flat_x = []
    flat_y = []
    for column, column_data in data.items():
        nan_free_column = column_data.dropna()
        if not np.all(nan_free_column == 0):
            flat_x += nan_free_column.size * [column]
            flat_y += list(nan_free_column)
    return flat_x, flat_y


def _subplot_layout_(plots, **kwargs) -> Dict[str, Any]:
    n_rows = _ceiling_division_(len(plots), N_COLS)
    return {
        'grid': {'rows': n_rows, 'columns': N_COLS, 'pattern': 'independent'},
        'annotations': [
            {
                'text': lclass,
                'showarrow': False,
                'x': 0,
                'xref': f'x{i + 1} domain',
                'y': 1.15,
                'yref': f'y{i + 1} domain',
                **kwargs
            }
            for i, lclass in enumerate(sorted(plots.keys()))
        ]
    }


def _df_to_plotly_dict_(
    data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
    plot_type: str = 'box', x_to_string: bool = True,
    y_to_string: bool = False
):
    if isinstance(data, pd.DataFrame):
        if 'Group' in data.keys():
            groups = data['Group']
            unique_groups = groups.unique()
            tmp_data = data.drop(columns='Group')
            colours = _get_colours_(unique_groups.size)
            plot_list = []
            for i, group in enumerate(sorted(unique_groups)):
                x, y = _flatten_(tmp_data.loc[groups == group, :])
                plot_list.append({
                    'x': [str(xi) for xi in x] if x_to_string else x,
                    'y': [str(yi) for yi in y] if y_to_string else y,
                    'name': group,
                    'type': plot_type,
                    'legendgroup': group,
                    'marker': {'color': colours[i]}
                })
        else:
            x, y = _flatten_(data)
            plot_list = [{
                    'x': [str(xi) for xi in x] if x_to_string else x,
                    'y': y,
                    'type': plot_type,
                    'marker': {'color': DEFAULT_COLOUR}
            }]
        return plot_list
    elif isinstance(data, dict):
        plotly_format = {
            lclass: _df_to_plotly_dict_(class_data)
            for lclass, class_data in sorted(data.items())
        }
        # computing layout
        layout = _subplot_layout_(plotly_format)
        # computing actual plots
        plotly_subplots = []
        for i, (lclass, plotly_data) in enumerate(plotly_format.items()):
            for tmp_data in plotly_data:
                tmp_data['xaxis'] = f'x{i + 1}'
                tmp_data['yaxis'] = f'y{i + 1}'
                tmp_data['title'] = lclass
                tmp_data['showlegend'] = i == 0
                plotly_subplots.append(tmp_data)
        return plotly_subplots, layout


def _df_to_plotly_heatmap_(
    data: Union[pd.DataFrame, np.ndarray],
    index: pd.Index = None, columns: pd.Index = None,
    scale: Tuple[int, int] = (5, 35),
    extrema: Tuple[float, float] = None, **kwargs
) -> Dict[str, Any]:
    zero_mask = data != 0
    if extrema is None:
        scaled_values = np.array(data) / (np.nanmax(data) - np.nanmin(data)[zero_mask])
    else:
        scaled_values = np.array(data) / (extrema[1] - extrema[0])
    scaled_values = scaled_values * (scale[1] - scale[0]) + scale[0]
    if isinstance(data, np.ndarray):
        if index is None:
            raise ValueError("'index' must be given, when numpy array is given")
        if columns is None:
            raise ValueError("'index' must be given, when numpy array is given")
        zero_mask = zero_mask.flatten()
        text = [f'Concentration: {value}' for i, value in enumerate(data.flatten())
                if zero_mask[i]]
    else:
        index = data.index
        columns = data.columns
        zero_mask = zero_mask.values.flatten()
        text = [f'Concentration: {value}' for i, value in enumerate(data.values.flatten())
                if zero_mask[i]]
    x, y = [], []
    x = data.shape[0] * list(columns)
    for i in range(data.shape[0]):
        y += data.shape[1] * [index[i]]
    scaled_values = list(scaled_values.flatten()[zero_mask])
    return {
        'x': [str(xi) for xi in np.array(x)[zero_mask]],
        'y': [str(yi) for yi in np.array(y)[zero_mask]],
        'marker': {
            'size': scaled_values,
            'color': kwargs.pop('color', DEFAULT_COLOUR),
            'line': {'width': 0}
        },
        'type': kwargs.pop('type', 'heatmap'),
        'text': text,
        **kwargs
    }


def _plotly_c_vs_db_(
    data: Dict[str, Union[Dict[str, pd.DataFrame], pd.DataFrame]],
    with_groups: bool = True, alpha: float = 1, **kwargs
) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
    if with_groups:
        min_ = np.nanmin([np.nanmin(x) for class_data in data.values() for x in class_data.values()])
        max_ = np.nanmax([np.nanmax(x) for class_data in data.values() for x in class_data.values()])
    else:
        min_ = np.nanmin([np.nanmin(x) for x in data.values()])
        max_ = np.nanmax([np.nanmax(x) for x in data.values()])
    plots = {}
    for i, (lclass, class_data) in enumerate(sorted(data.items())):
        if with_groups:
            colours = _get_colours_(len(class_data))
            for j, (group, group_data) in enumerate(sorted(class_data.items())):
                # TODO: add scatter offset per group
                pt_form = _df_to_plotly_heatmap_(
                    group_data, type='scatter',
                    color=colours[j],
                    legendgroup=group, extrema=(min_, max_),
                    mode='markers', name=group,
                    opacity=alpha,
                    # boxpoints='all',
                    # hoveron='points', pointpos=0,
                    # boxmode='group', line={'color': 'rgba(0,0,0,0)'},
                    # fillcolor='rgba(0,0,0,0)'
                )
                pt_form['xaxis'] = f'x{i + 1}'
                pt_form['yaxis'] = f'y{i + 1}'
                if i == 0:
                    pt_form['title'] = lclass
                    pt_form['showlegend'] = True
                else:
                    pt_form['showlegend'] = False
                plots.setdefault(lclass, []).append(pt_form)
        else:
            plots[lclass] = [
                _df_to_plotly_heatmap_(class_data, type='scatter', extrema=(min_, max_))
            ]
    return [plot_data for plot in plots.values() for plot_data in plot], _subplot_layout_(plots, **kwargs)


def _per_sample_sums_(
    data: pd.DataFrame, lipid_map: Dict[str, List[Lipid]],
    groups: pd.Series = None
) -> Dict[str, Union[Dict[str, pd.Series], Dict[str, Dict[str, pd.Series]]]]:
    # data: lipids in columns
    classes = {}
    lengths = {}
    dbs = {}
    for lipid_class, lipids in lipid_map.items():
        classes[lipid_class] = [lipid.get_dataname() for lipid in lipids]
        for lipid in lipids:
            lengths.setdefault(lipid.sum_length(), []).append(lipid.get_dataname())
            dbs.setdefault(lipid.sum_dbs(), []).append(lipid.get_dataname())
    class_data = {lclass: data.loc[lipids, :] for lclass, lipids in classes.items()}
    sums = {
        'Lipid Class': {},
        'Sum Length': {},
        'Sum DB': {},
        'C vs DB': {}
    }
    for lclass, lclass_data in class_data.items():
        sums['Lipid Class'][lclass] = lclass_data.sum(axis=0)
        sums['Sum Length'][lclass] = {
            llength: lclass_data.loc[set(lipids).intersection(lclass_data.index), :].sum(axis=0)
            for llength, lipids in lengths.items()
        }
        sums['Sum DB'][lclass] = {
            ldb: lclass_data.loc[set(lipids).intersection(lclass_data.index), :].sum(axis=0)
            for ldb, lipids in dbs.items()
        }
        if groups is None:
            sums['C vs DB'][lclass] = {
                llength: {
                    ldb: lclass_data.loc[
                            set(c_lipids).intersection(db_lipids).intersection(lclass_data.index), :
                         ].sum(axis=0).mean()
                    for ldb, db_lipids in dbs.items()
                } for llength, c_lipids in lengths.items()
            }
        else:
            unique_groups = groups.unique()
            sums['C vs DB'][lclass] = {
                group: {
                    llength: {
                        ldb: lclass_data.loc[
                                set(c_lipids).intersection(db_lipids).intersection(lclass_data.index), groups == group
                             ].sum(axis=0).mean()
                        for ldb, db_lipids in dbs.items()
                    } for llength, c_lipids in lengths.items()
                } for group in unique_groups
            }
    return sums


def lipidome_summary(
    self, as_dict: bool = False, **kwargs
) -> Union[Tuple[pd.DataFrame, dict, dict, dict], Tuple[list, tuple, tuple, tuple]]:
    """
    TODO
    :param self:
    :param as_dict:
    :return:
    """
    sum_data = _per_sample_sums_(self.data, self.gln.get_lipid_dict(), self.groups)
    class_sums = pd.DataFrame(sum_data['Lipid Class'])
    # normalise by total class content
    length_summary = {
        lclass: (pd.DataFrame(sum_length).T / class_sums[lclass]).T
        for lclass, sum_length in sum_data['Sum Length'].items()
    }
    db_summary = {
        lclass: (pd.DataFrame(sum_db).T / class_sums[lclass]).T
        for lclass, sum_db in sum_data['Sum DB'].items()
    }
    class_summary = (class_sums.T / class_sums.sum(axis=1)).T

    if self.groups is not None:
        # Group assignment
        class_summary['Group'] = self.groups
        length_summary = {lclass: sum_length.assign(Group=self.groups)
                          for lclass, sum_length in length_summary.items()}
        # x/y-axis: C/DB, value: (per group) mean
        class_sums['Group'] = self.groups
        class_means = class_sums.groupby('Group').mean()
        c_db_scatter = {
            lclass: {
                group: pd.DataFrame.from_dict(group_data) / class_means.loc[group, lclass]
                for group, group_data in c_db.items()
            } for lclass, c_db in sum_data['C vs DB'].items()
        }
        db_summary = {lclass: db_length.assign(Group=self.groups)
                      for lclass, db_length in db_summary.items()}
        with_groups = True
    else:
        with_groups = False
        # x/y-axis: C/DB, value: mean
        class_means = class_sums.mean(axis=0, skipna=True)
        c_db_scatter = {lclass: pd.DataFrame.from_dict(c_db) / class_means[lclass]
                        for lclass, c_db in sum_data['C vs DB'].items()}
    if as_dict:
        return _df_to_plotly_dict_(class_summary), \
               _df_to_plotly_dict_(length_summary), \
               _df_to_plotly_dict_(db_summary), \
               _plotly_c_vs_db_(c_db_scatter, with_groups, xaxis={'title': 'Sum Length'},
                                yaxis={'title': 'Sum DBs'}, **kwargs)
    return class_summary, length_summary, db_summary, c_db_scatter


def _z_score_(data: np.ndarray) -> np.ndarray:
    return ((data.T - np.nanmean(data, axis=1)) / np.nanstd(data, axis=1)).T


def substructure_selection(
    self, functions: list, **kwargs
) -> Tuple[pd.DataFrame, Union[pd.Series, pd.DataFrame]]:
    """
    TODO
    :param self:
    :param functions:
    :param kwargs:
    :return:
    """
    if self.groups is None:
        # TODO unsupervised group assignment
        raise ValueError('groups have to be given for feature selection')
    ohencoded = False
    if np.issubdtype(self.groups.dtype, np.number):
        groups = self.groups
    else:
        ohencoded = True
        groups = pd.get_dummies(self.groups)
    substructure_data = analysis_lipid_substructure(
        self, functions,
        single_n_combination=kwargs.pop('single_n_combionations', True),
        feature_combinations=kwargs.pop('feature_combinations', len(functions)),
        **kwargs
    )
    substructure_data = pd.DataFrame(substructure_data.values,
                                     index=substructure_data.index,
                                     columns=substructure_data.columns)
    # TODO: elastic net or similar?
    model = LinearRegression()
    substructs = substructure_data.replace([np.inf, -np.inf], np.nan).fillna(0)
    model.fit(substructs.values.T, groups.values)
    # TODO: add some prediction plot?
    if isinstance(groups, pd.Series):
        return substructure_data, pd.Series(model.coef_, substructure_data.index).sort_values(ascending=False)
    else:
        return substructure_data, pd.DataFrame(model.coef_, index=groups.columns, columns=substructure_data.index)


def plot_pca(
    data, groups: pd.Series = None,
    to_plotly: bool = False, title: str = '',
    ax: Optional[plt.axis] = None, scale: bool = True
) -> Union[Tuple[List[Dict[str, Dict[str, Any]]], Dict[str, Any]], plt.axis]:
    """
    TODO
    :param data:
    :param groups:
    :param to_plotly:
    :param title:
    :param ax:
    :param scale:
    :return:
    """
    pca = PCA(n_components=2)
    is_df = False
    if isinstance(data, pd.DataFrame):
        is_df = True
        if scale:
            scaled = _z_score_(data.values)
            scaled[np.isnan(scaled)] = 0
            scaled[np.isinf(data)] = 0
            embedding = pca.fit_transform(scaled.T)
        else:
            embedding = pca.fit_transform(data.replace([np.inf, -np.inf], np.nan).fillna(0).T)
    else:
        data_ = data.copy()
        data_[np.isinf(data)] = 0
        data_[np.isnan(data)] = 0
        if scale:
            data_ = _z_score_(data_)
        embedding = pca.fit_transform(data_.T)
    if to_plotly:
        layout = {
            'xaxis': {'title': f'PC1 ({round(100 * pca.explained_variance_ratio_[0], ndigits=2)}%)'},
            'yaxis': {'title': f'PC2 ({round(100 * pca.explained_variance_ratio_[1], ndigits=2)}%)'},
            'title': title
        }
        if groups is not None:
            unique_groups = groups.unique()
            colours = _get_colours_(unique_groups.size)
            plot_data = [
                {
                    'x': list(embedding[groups[data.columns] == group, 0]),
                    'y': list(embedding[groups[data.columns] == group, 1]),
                    'type': 'scatter',
                    'name': group,
                    'mode': 'markers',
                    'text': data.columns.to_list() if is_df else [],
                    'markers': {'color': colours[i]}
                }
                for i, group in enumerate(sorted(unique_groups))
            ]
        else:
            plot_data = [
                {
                    'x': list(embedding[:, 0]),
                    'y': list(embedding[:, 1]),
                    'type': 'scatter',
                    'mode': 'markers',
                    'text': data.columns.to_list() if is_df else []
                }
            ]
        return plot_data, layout
    # TODO: plot with matplotlib
    if ax is None:
        fig, ax = plt.subplots()
    if groups is not None:
        unique_groups = groups.unique()
        colours = _get_colours_(unique_groups.size)
        for i, group in enumerate(sorted(unique_groups)):
            ax.scatter(
                embedding[groups[data.columns] == group, 0],
                embedding[groups[data.columns] == group, 1],
                label=group, c=colours[i]
            )
    else:
        ax.scatter(
            embedding[:, 0],
            embedding[:, 1]
        )
    ax.set_xlabel(f'PC1 ({round(100 * pca.explained_variance_ratio_[0], ndigits=2)}%)')
    ax.set_ylabel(f'PC2 ({round(100 * pca.explained_variance_ratio_[1], ndigits=2)}%)')
    return ax


def plot_heatmap(
    data, scale: bool = True,
    groups: pd.Series = None,
    to_plotly: bool = False, title: str = '',
    ax: Optional[plt.axis] = None
) -> Union[Dict[str, Any], sns.matrix.ClusterGrid]:
    """
    TODO
    :param data:
    :param scale:
    :param groups:
    :param to_plotly:
    :param title:
    :param ax:
    :return:
    """
    if scale:
        plot_data = _scale_(data.values)
    else:
        plot_data = data.values
    if groups is not None:
        unique_groups = groups.unique()
        group_colours = dict(zip(unique_groups, _get_colours_(unique_groups.size)))
        column_colours = groups[plot_data.columns].map(group_colours)
    if to_plotly:
        raise NotImplementedError("'to_plotly' currently not implemented for 'plot_heatmap'")
        # TODO: dendrogram/clustering order (columns) and column annotation
        # if groups is not None:
        #     return _df_to_plotly_heatmap_(plot_data, data.index, data.columns, title=title)
        # else:
        #     return _df_to_plotly_heatmap_(plot_data, data.index, data.columns, title=title)
    plot_data = pd.DataFrame(plot_data, index=data.index, columns=data.columns)
    if groups is not None:
        return sns.clustermap(plot_data, column_colors=column_colours, ax=ax)
    else:
        return sns.clustermap(plot_data, ax=ax)



from typing import List, Dict
from .reaction import Reaction, FAReaction
from .lipid import Lipid, FA
from .edgelist import Edge, Edgelist
from .reaction_types import *
from .reference import ReferenceLipid
from multiprocessing import Process, Queue
from datetime import datetime
import itertools

class Extender:

    __slots__ = ["__class_reacs", "__fa_reacs"]

    def __init__(self, class_reactions: List[Reaction], fatty_acid_reactions: List[FAReaction]):
        self.__class_reacs = class_reactions
        self.__fa_reacs = fatty_acid_reactions

    def evaluate_fa_reactions(self, lip_dict: Dict[str, List[Lipid]],
                              excluded_fa_conversions: List[List[FA]] = None, startid=0) -> Edgelist:

        return evaluate_fa_reactions_static(lip_dict, self.__fa_reacs, excluded_fa_conversions, startid=startid)

    def evaluate_fa_reactions_parallel(self, lip_dict: Dict[str, List[Lipid]], n_processes: int,
                              excluded_fa_conversions: List[List[FA]] = None) -> Edgelist:
        queue = Queue()
        split = len(self.__class_reacs) // n_processes
        p = [Process(target=evaluate_fa_reactions_worker,
                     args=(queue, lip_dict, self.__fa_reacs[(split * x):(split * (x + 1))], excluded_fa_conversions)) for x in
             range(n_processes - 1)]
        p.append(Process(target=evaluate_fa_reactions_worker,
                         args=(queue, lip_dict, self.__fa_reacs[(split * (n_processes - 1)):], excluded_fa_conversions)))

        for pi in p:
            pi.start()
        results = [queue.get() for pr in range(n_processes)]
        for pj in p:
            pj.join()

        a = Edgelist()
        for r in results:
            a = a + r

        return a

    def evaluate_class_reactions(self, lip_dict: Dict[str, List[Lipid]], startid: int =0) -> Edgelist:
        return evaluate_class_reactions_static(self.__class_reacs, lip_dict, startid=startid)

    def evaluate_ether_conversions(self, lip_dict: Dict[str, List[Lipid]],
                                   reference_lipids: List[ReferenceLipid], startid: int =0) -> Edgelist:
        return evaluate_ether_conversions_static(lip_dict, reference_lipids=reference_lipids, startid=startid)

    def evaluate_class_reactions_parallel(self, lip_dict: Dict[str, List[Lipid]], n_processes: int) -> Edgelist:
        queue = Queue()
        split = len(self.__class_reacs) // n_processes
        p = [Process(target=evaluate_class_reactions_worker,
                     args=(queue, self.__class_reacs[(split*x):(split*(x+1))], lip_dict)) for x in range(n_processes-1)]
        p.append(Process(target=evaluate_class_reactions_worker,
                         args=(queue, self.__class_reacs[(split*(n_processes-1)):], lip_dict)))
        # p1 = Process(target=evaluate_class_reactions_worker, args=(queue, self.__class_reacs[:split], lip_dict))
        # p2 = Process(target=evaluate_class_reactions_worker, args=(queue, self.__class_reacs[split:], lip_dict))
        # p1.start()
        # p2.start()
        for pi in p:
            pi.start()
        results = [queue.get() for pr in range(n_processes)]
        for pj in p:
            pj.join()

        a = Edgelist()
        for r in results:
            a = a + r

        return a

    def evaluate_all(self, lip_dict: Dict[str, List[Lipid]],
                     excluded_fa_conversions: List[List[FA]] = None,
                     ether_conversions: bool = False,
                     reference_lipids: List[ReferenceLipid] = None) -> Edgelist:
        # Class reactions
        tmp1 = self.evaluate_class_reactions(lip_dict)
        tmp1_ed = tmp1.raw_edgelist_copy()
        if len(tmp1_ed) > 0:
            stid = tmp1.raw_edgelist_copy()[-1].get_reaction_id()
        else:
            stid = 0
        # stid = tmp1.raw_edgelist_copy()[-1].get_reaction_id()
        # FA reactions
        tmp2 = self.evaluate_fa_reactions(lip_dict, excluded_fa_conversions, startid=int(stid)+1)
        tmp2_ed = tmp2.raw_edgelist_copy()
        if len(tmp2_ed) > 0:
            stid = tmp2.raw_edgelist_copy()[-1].get_reaction_id()
        # Ether conversions
        tmp3 = Edgelist()
        if ether_conversions:
            if reference_lipids is not None:
                tmp3 = self.evaluate_ether_conversions(lip_dict=lip_dict,
                                                       reference_lipids=reference_lipids,
                                                       startid=int(stid)+1)

        return tmp1 + tmp2 + tmp3

    def evaluate_all_parallel(self, lip_dict: Dict[str, List[Lipid]], n_processes: int,
                              excluded_fa_conversions: List[List[FA]] = None) -> Edgelist:
        tmp1 = self.evaluate_class_reactions(lip_dict)
        tmp2 = self.evaluate_fa_reactions_parallel(lip_dict, n_processes, excluded_fa_conversions)
        return tmp1 + tmp2


def lfad22_inner(i: Reaction, r1: List[Lipid], l1: Lipid, l2: Lipid, l3: Lipid, rid: int = 0) -> List[Edge]:
    for l4 in r1:
        fa_set = set(l1.get_fas() + l2.get_fas() + l3.get_fas() + l4.get_fas())
        for fa in fa_set:
            if (sorted(l1.get_fas()) == sorted(l2.get_fas() + [fa]) or
                sorted(l1.get_fas() + [fa]) == sorted(l2.get_fas())) and \
                    (sorted(l3.get_fas()) == sorted(l4.get_fas() + [fa]) or
                     sorted(l3.get_fas() + [fa]) == sorted(l4.get_fas())):

                return [Edge(l1, l2, enzyme_id=i.get_enzyme_id(), reaction_id=str(rid),
                             reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                             gene_name=i.get_gene_name(),
                             reaction_structure="2,2", l1_type="substrate", l2_type="product",
                             nl_participants=i.get_nl_participants()),
                        Edge(l3, l4, enzyme_id=i.get_enzyme_id(), reaction_id=str(rid),
                             reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                             gene_name=i.get_gene_name(),
                             reaction_structure="2,2", l1_type="substrate", l2_type="product",
                             nl_participants=i.get_nl_participants())]
    return []


def lhg22_inner(i: Reaction, r1: List[Lipid], l1: Lipid, l2: Lipid, l3: Lipid) -> List[Edge]:
    for l4 in r1:
        if sorted(l1.get_fas()) == sorted(l2.get_fas()) and \
                sorted(l3.get_fas()) == sorted(l4.get_fas()):
            return [Edge(l1, l2, enzyme_id=i.get_enzyme_id(), reaction_id="r_id",
                         reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                             gene_name=i.get_gene_name(),
                             nl_participants=i.get_nl_participants()),
                    Edge(l3, l4, enzyme_id=i.get_enzyme_id(), reaction_id="r_id",
                         reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                             gene_name=i.get_gene_name(),
                             nl_participants=i.get_nl_participants())]
    return []


def evaluate_class_reactions_static(class_reacs: List[Reaction],
                                    lip_dict: Dict[str, List[Lipid]], startid: int = 0) -> Edgelist:
    edge_list = Edgelist()

    # TODO: Function too big. Should be split up into multiple functions.

    # Loop over class reactions
    curr_reac_id = startid
    l_counter = 0
    for i in class_reacs:
        # print(f"reaction {l_counter} of {len(class_reacs)} ({str(i)})")
        # l_counter += 1

        if i.is_valid(verbose=False):
            if i.subs_eq_prods():
                continue
            if i.shape() == (0, 1) or i.shape() == (1, 0):
                pass
            elif i.shape() == (1, 1):
                # Pick both classed from lip_dict
                sub_ref = i.get_substrates()[0]
                prod_ref = i.get_products()[0]
                if sub_ref.get_abbr() in lip_dict and prod_ref.get_abbr() in lip_dict:
                    pot_subs = lip_dict[sub_ref.get_abbr()]
                    pot_prods = lip_dict[prod_ref.get_abbr()]

                    if i.get_reaction_type() == L_HGmodify or i.get_reaction_type() == L_HGdelete:
                        # Loop over both of them
                        for l1 in pot_subs:
                            for l2 in pot_prods:
                                # get pairs
                                if sorted(l1.get_fas()) == sorted(l2.get_fas()):
                                    edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                                                            reaction_id=str(curr_reac_id),
                                                            reaction_type=i.get_reaction_type(), notes="",
                                                            uniprot=i.get_uniprot(),
                                                            gene_name=i.get_gene_name(),
                                                            fa_param=str(l1.sum_length()),
                                                            nl_participants=i.get_nl_participants()))
                                    curr_reac_id += 1
                                    break
                    elif i.get_reaction_type() == L_FAdelete:
                        # Loop over both of them
                        for l1 in pot_subs:
                            for l2 in pot_prods:
                                # get pairs
                                if l1.get_number_fas() < l2.get_number_fas():
                                    f1 = l1.get_fas()
                                    f2 = l2.get_fas()
                                else:
                                    f2 = l1.get_fas()
                                    f1 = l2.get_fas()
                                for ff2 in range(len(f2)):
                                    if sorted(f2) == sorted(f1 + [f2[ff2]]):
                                        edge_list.add_edge(
                                            Edge(l1, l2, enzyme_id=i.get_enzyme_id(), reaction_id=str(curr_reac_id),
                                                 reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                                                 gene_name=i.get_gene_name(),
                                                 fa_param=str(f2[ff2]),
                                                 nl_participants=i.get_nl_participants()))
                                        curr_reac_id += 1
                                        break
                    # This is only for reaction transforming lipids into ether lipids
                    elif i.get_reaction_type() == L_FAether:
                        sub_ref = i.get_substrates()[0]
                        prod_ref = i.get_products()[0]

                        if (sub_ref.get_headgroup() == prod_ref.get_headgroup()) and \
                            (sub_ref.get_fas() == prod_ref.get_fas()) and \
                                (sub_ref.get_lcb() == prod_ref.get_lcb()):
                            if (sub_ref.get_ether() == (prod_ref.get_ether()+1)) or \
                                    (prod_ref.get_ether() == (sub_ref.get_ether() + 1)):
                                for l1 in pot_subs:
                                    for l2 in pot_prods:
                                        if sorted([x.as_not_ether() for x in l1.get_fas()]) == \
                                            sorted([x.as_not_ether() for x in l2.get_fas()]):
                                            edge_list.add_edge(
                                                Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                                                     reaction_id=str(curr_reac_id),
                                                     reaction_type=i.get_reaction_type(),
                                                     notes="",
                                                     uniprot=i.get_uniprot(),
                                                     gene_name=i.get_gene_name(),
                                                     fa_param=l1.sum_length(),
                                                     nl_participants=i.get_nl_participants()))
                                            curr_reac_id += 1
                                            break



            elif i.shape() == (1, 2) or i.shape() == (2, 1):
                if i.get_reaction_type() == L_FAdelete:
                    twos = i.get_substrates() if len(i.get_substrates()) == 2 else i.get_products()
                    ones = i.get_substrates() if len(i.get_substrates()) == 1 else i.get_products()
                    one = ones[0]
                    t1 = twos[0]
                    t2 = twos[1]
                    # find partner
                    if (one.get_headgroup() == t1.get_headgroup()) and (one.get_lcb() == t1.get_lcb()) and \
                            (one.get_ether() == t1.get_ether()) and (one.get_fas() == (t1.get_fas()+1)):
                        partner = t1
                        other = t2
                    else:
                        partner = t2
                        other = t1
                    if one.get_abbr() in lip_dict and partner.get_abbr() in lip_dict and \
                            other.get_abbr() in lip_dict:
                        one_pos = lip_dict[one.get_abbr()]
                        partner_pot = lip_dict[partner.get_abbr()]
                        other_pot = lip_dict[other.get_abbr()]
                        for l1 in one_pos:
                            for l2 in partner_pot:
                                for l3 in other_pot:
                                    for fa in l3.get_fas():
                                        if sorted(l1.get_fas()) == sorted(l2.get_fas() + [fa]):
                                            edge_list.add_edges([Edge(l1, l2,
                                                                      enzyme_id=i.get_enzyme_id(),
                                                                      reaction_id=str(curr_reac_id),
                                                                      reaction_type=i.get_reaction_type(),
                                                                      notes="", uniprot=i.get_uniprot(),
                                                                      gene_name=i.get_gene_name(),
                                                                      reaction_structure="1,2",
                                                                      l1_type="substrate", l2_type="product",
                                                                      fa_param=str(fa),
                                                                      nl_participants=i.get_nl_participants()),
                                                                 # Makes only sense in bipartite networks...
                                                                 Edge(l1, l3,
                                                                      enzyme_id=i.get_enzyme_id(),
                                                                      reaction_id=str(curr_reac_id),
                                                                      reaction_type=i.get_reaction_type(),
                                                                      notes="", uniprot=i.get_uniprot(),
                                                                      gene_name=i.get_gene_name(),
                                                                      reaction_structure="1,2",
                                                                      l1_type="substrate", l2_type="product",
                                                                      fa_param=str(fa),
                                                                      nl_participants=i.get_nl_participants())
                                                                 ]
                                                )

                                            curr_reac_id += 1
                                            break
                elif i.get_reaction_type() == L_merge:
                    twos = i.get_substrates() if len(i.get_substrates()) == 2 else i.get_products()
                    ones = i.get_substrates() if len(i.get_substrates()) == 1 else i.get_products()
                    one = ones[0]
                    t1 = twos[0]
                    t2 = twos[1]
                    if one.get_abbr() in lip_dict and t1.get_abbr() in lip_dict and t2.get_abbr() in lip_dict:
                        one_pos = lip_dict[one.get_abbr()]
                        partner_pot = lip_dict[t1.get_abbr()]
                        other_pot = lip_dict[t2.get_abbr()]

                        for l1 in one_pos:
                            for l2 in partner_pot:
                                for l3 in other_pot:
                                    if sorted(l1.get_fas()) == sorted(l2.get_fas() + l3.get_fas()):
                                        edge_list.add_edge(
                                            Edge(l1, l2, enzyme_id=i.get_enzyme_id(), reaction_id=str(curr_reac_id),
                                                 reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                                                 gene_name=i.get_gene_name(),
                                                 reaction_structure="1,2",
                                                 l1_type="substrate", l2_type="product",
                                                 fa_param=str(l2.sum_length()),
                                                 nl_participants=i.get_nl_participants()))
                                        edge_list.add_edge(
                                            Edge(l1, l3, enzyme_id=i.get_enzyme_id(), reaction_id=str(curr_reac_id),
                                                 reaction_type=i.get_reaction_type(), notes="", uniprot=i.get_uniprot(),
                                                 gene_name=i.get_gene_name(),
                                                 reaction_structure="1,2",
                                                 l1_type="substrate", l2_type="product",
                                                 fa_param=str(l2.sum_length()),
                                                 nl_participants=i.get_nl_participants()))
                                        curr_reac_id += 1
                                        break
            elif i.shape() == (2, 2):
                subs = i.get_substrates()
                prods = i.get_products()
                if i.get_reaction_type() == L_FAdelete:
                    if subs[0].get_headgroup() == prods[0].get_headgroup() and \
                            subs[0].get_lcb() == prods[0].get_lcb() and \
                            subs[0].get_ether() == prods[0].get_ether() and \
                            subs[0].get_potentialfas() == prods[0].get_potentialfas() and \
                            ((subs[0].get_fas() == (prods[0].get_fas()+1) or
                              (prods[0].get_fas() == (subs[0].get_fas()+1)))):
                        p11 = subs[0]
                        p12 = prods[0]
                        p21 = subs[1]
                        p22 = prods[1]
                    else:
                        p11 = subs[0]
                        p12 = prods[1]
                        p21 = subs[1]
                        p22 = prods[0]

                    # get higher partner p1
                    if p11.get_fas() > p12.get_fas():
                        higher_p1 = p11
                        lower_p1 = p12
                    else:
                        higher_p1 = p12
                        lower_p1 = p11

                    #get higher partner p2
                    if p21.get_fas() > p22.get_fas():
                        higher_p2 = p21
                        lower_p2 = p22
                    else:
                        higher_p2 = p22
                        lower_p2 = p21


                    if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and \
                            p21.get_abbr() in lip_dict and p22.get_abbr() in lip_dict:

                        # Uncoupling 2,2 L_FAdelete reactions for local search

                        for l1 in lip_dict[higher_p1.get_abbr()]:
                            for l2 in lip_dict[lower_p1.get_abbr()]:
                                for c1 in itertools.combinations(l1.get_fas(), l1.get_number_fas()-1):
                                    if sorted(list(c1)) == sorted(l2.get_fas()):
                                        for fa1 in l1.get_fas():
                                            if sorted([fa1] + list(c1)) == sorted(l1.get_fas()):
                                                edge_list.add_edges([Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                                                                          reaction_id=str(curr_reac_id),
                                                                          reaction_type=i.get_reaction_type(),
                                                                          notes="", uniprot=i.get_uniprot(),
                                                                          gene_name=i.get_gene_name(),
                                                                          reaction_structure="2,2",
                                                                          l1_type="substrate",
                                                                          l2_type="product",
                                                                          fa_param=str(fa1),
                                                                          nl_participants=i.get_nl_participants()
                                                                          )])
                                                curr_reac_id += 1
                                    break

                        for l1 in lip_dict[higher_p2.get_abbr()]:
                            for l2 in lip_dict[lower_p2.get_abbr()]:
                                for c1 in itertools.combinations(l1.get_fas(), l1.get_number_fas()-1):
                                    if sorted(list(c1)) == sorted(l2.get_fas()):
                                        for fa1 in l1.get_fas():
                                            if sorted([fa1] + list(c1)) == sorted(l1.get_fas()):
                                                edge_list.add_edges([Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                                                                          reaction_id=str(curr_reac_id),
                                                                          reaction_type=i.get_reaction_type(),
                                                                          notes="", uniprot=i.get_uniprot(),
                                                                          gene_name=i.get_gene_name(),
                                                                          reaction_structure="2,2",
                                                                          l1_type="substrate",
                                                                          l2_type="product",
                                                                          fa_param=str(fa1),
                                                                          nl_participants=i.get_nl_participants()
                                                                          )])
                                                curr_reac_id += 1
                                    break

                        # for l1 in lip_dict[higher_p1.get_abbr()]:
                        #     for l2 in lip_dict[lower_p1.get_abbr()]:
                        #         for c1 in itertools.combinations(l1.get_fas(), l1.get_number_fas()-1):
                        #             if sorted(list(c1)) == sorted(l2.get_fas()):
                        #                 for fa1 in l1.get_fas():
                        #                     if sorted([fa1] + list(c1)) == sorted(l1.get_fas()):
                        #                         for l3 in lip_dict[higher_p2.get_abbr()]:
                        #                             if fa1 in l3.get_fas():
                        #                                 for l4 in lip_dict[lower_p2.get_abbr()]:
                        #                                     fa_set = set(
                        #                                         l1.get_fas() + l2.get_fas() + l3.get_fas() + l4.get_fas())
                        #                                     for fa in fa_set:
                        #                                         if (sorted(l1.get_fas()) == sorted(
                        #                                                 l2.get_fas() + [fa]) or
                        #                                             sorted(l1.get_fas() + [fa]) == sorted(
                        #                                                     l2.get_fas())) and \
                        #                                                 (sorted(l3.get_fas()) == sorted(
                        #                                                     l4.get_fas() + [fa]) or
                        #                                                  sorted(l3.get_fas() + [fa]) == sorted(
                        #                                                             l4.get_fas())):
                        #
                        #                                             edge_list.add_edges([Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                        #                                                          reaction_id=str(curr_reac_id),
                        #                                                          reaction_type=i.get_reaction_type(),
                        #                                                          notes="", uniprot=i.get_uniprot(),
                        #                                                          gene_name=i.get_gene_name(),
                        #                                                          reaction_structure="2,2",
                        #                                                          l1_type="substrate",
                        #                                                          l2_type="product",
                        #                                                          fa_param=str(fa1),
                        #                                                          nl_participants=i.get_nl_participants()
                        #                                                                       ),
                        #                                                     Edge(l3, l4, enzyme_id=i.get_enzyme_id(),
                        #                                                          reaction_id=str(curr_reac_id),
                        #                                                          reaction_type=i.get_reaction_type(),
                        #                                                          notes="", uniprot=i.get_uniprot(),
                        #                                                          gene_name=i.get_gene_name(),
                        #                                                          reaction_structure="2,2",
                        #                                                          l1_type="substrate",
                        #                                                          l2_type="product",
                        #                                                          fa_param=str(fa1),
                        #                                                          nl_participants=i.get_nl_participants()
                        #                                                          )
                        #                                                                  ])
                        #                                             curr_reac_id += 1
                        #
                        #
                        #                         break

                    # if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and \
                    #         p21.get_abbr() in lip_dict and p22.get_abbr() in lip_dict:
                    #     for l1 in lip_dict[p11.get_abbr()]:
                    #         for l2 in lip_dict[p12.get_abbr()]:
                    #
                    #             for l3 in lip_dict[p21.get_abbr()]:
                    #                 edge_list.add_edges(lfad22_inner(i, lip_dict[p22.get_abbr()], l1, l2, l3, rid=curr_reac_id))
                    #                 curr_reac_id += 1

                elif i.get_reaction_type() == L_HGdelete or i.get_reaction_type() == L_HGmodify:
                    # TODO: Currently not working for complex L_HGmodify reactions,
                    #  e.g. parts of an headgroup is added to another one
                    if subs[0].get_headgroup() == prods[0].get_headgroup() and \
                            subs[0].get_lcb() == prods[1].get_lcb() and subs[0].get_ether() == prods[1].get_ether():
                        p11 = subs[0] # partner: (p11, p12) and (p21, p22)
                        p12 = prods[1]
                        p21 = subs[1]
                        p22 = prods[0]
                    else:
                        p11 = subs[0]
                        p12 = prods[0]
                        p21 = subs[1]
                        p22 = prods[1]
                    if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and \
                            p21.get_abbr() in lip_dict and p22.get_abbr() in lip_dict:
                        # Evaluate first partner set (p11, p12):
                        for l1 in lip_dict[p11.get_abbr()]:
                            for l2 in lip_dict[p12.get_abbr()]:
                                if sorted(l1.get_fas()) == sorted(l2.get_fas()):
                                    edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_id(),
                                                            reaction_id=str(curr_reac_id),
                                                            reaction_type=i.get_reaction_type(), notes="",
                                                            uniprot=i.get_uniprot(),
                                                            gene_name=i.get_gene_name(),
                                                            reaction_structure="2,2",
                                                            l1_type="substrate",
                                                            l2_type="product",
                                                            set_id=1,
                                                            fa_param=str(l1.sum_length()+l2.sum_length()),
                                                            nl_participants=i.get_nl_participants())
                                                            )

                                    curr_reac_id += 1
                                    break
                        # Evaluate second partner set (p21, p22):
                        for l3 in lip_dict[p21.get_abbr()]:
                            for l4 in lip_dict[p22.get_abbr()]:
                                if sorted(l3.get_fas()) == sorted(l4.get_fas()):
                                    edge_list.add_edge(Edge(l3, l4, enzyme_id=i.get_enzyme_id(),
                                                            reaction_id=str(curr_reac_id),
                                                            reaction_type=i.get_reaction_type(),
                                                            notes="", uniprot=i.get_uniprot(),
                                                            gene_name=i.get_gene_name(),
                                                            reaction_structure="2,2",
                                                            l1_type="substrate",
                                                            l2_type="product",
                                                            set_id=2,
                                                            fa_param=str(l3.sum_length()+l4.sum_length()),
                                                            nl_participants=i.get_nl_participants()))

                                    curr_reac_id += 1
                                    break

                    # old:
                    # if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and \
                    #         p21.get_abbr() in lip_dict and p22.get_abbr() in lip_dict:
                    #     for l1 in lip_dict[p11.get_abbr()]:
                    #         for l2 in lip_dict[p12.get_abbr()]:
                    #             for l3 in lip_dict[p21.get_abbr()]:
                    #                 edge_list.add_edges(lhg22_inner(i, lip_dict[p22.get_abbr()], l1, l2, l3))
    return edge_list

def evaluate_fa_reactions_static(lip_dict: Dict[str, List[Lipid]],
                                 fa_reacs: List[FAReaction],
                                 excluded_fa_conversions: List[List[FA]] = None, startid=0) -> Edgelist:
    curr_reac_id = startid
    edge_list = Edgelist()
    for far in fa_reacs:
        for lip in lip_dict.keys():
            if len(lip_dict[lip]) > 1:
                ll = lip_dict[lip]
                for l1 in range(len(ll) - 1):
                    for l2 in range(l1 + 1, len(ll)):
                        if far.is_transformable(ll[l1], ll[l2], excluded_reactions=excluded_fa_conversions):
                            edge_list.add_edge(Edge(ll[l1], ll[l2], enzyme_id="", reaction_id=str(curr_reac_id),
                                                    reaction_type=L_FAmodify, notes=far.get_last_conversion()))
                            curr_reac_id += 1
    return edge_list


def evaluate_class_reactions_worker(queue: Queue, class_reacs: List[Reaction],
                                    lip_dict: Dict[str, List[Lipid]]) -> int:
    queue.put(evaluate_class_reactions_static(class_reacs, lip_dict))
    return 0


def evaluate_fa_reactions_worker(queue: Queue, lip_dict: Dict[str, List[Lipid]],
                                 fa_reacs: List[FAReaction],
                                 excluded_fa_conversions: List[List[FA]] = None) -> int:
    queue.put(evaluate_fa_reactions_static(lip_dict, fa_reacs, excluded_fa_conversions))
    return 0

def evaluate_ether_conversions_static(lip_dict: Dict[str, List[Lipid]],
                                     reference_lipids: List[ReferenceLipid],
                                     startid: int = 0) -> Edgelist:
    curr_reac_id = startid
    edge_list = Edgelist()
    for i in range(len(reference_lipids)):
        for j in range(i+1, len(reference_lipids)):
            sub = reference_lipids[i]
            prod = reference_lipids[j]
            if sub.get_abbr() in lip_dict.keys() and prod.get_abbr() in lip_dict.keys():
                if (((sub.get_ether()-1) == prod.get_ether())
                        or (sub.get_ether() == (prod.get_ether()))-1) \
                       and (sub.get_lcb() == prod.get_lcb()) \
                       and (sub.get_potentialfas() == prod.get_potentialfas()) \
                       and (sub.get_headgroup() == prod.get_headgroup()) \
                       and ((sub.get_fas()) == prod.get_fas()):
                    for l1 in lip_dict[sub.get_abbr()]:
                        for l2 in lip_dict[prod.get_abbr()]:
                            if sorted([x.as_not_ether() for x in l1.get_fas()]) == \
                                    sorted([x.as_not_ether() for x in l2.get_fas()]):
                                edge_list.add_edge(
                                    Edge(l1, l2, enzyme_id="",
                                         reaction_id=str(curr_reac_id),
                                         reaction_type=L_FAmodify,
                                         notes="",
                                         uniprot="",
                                         gene_name="",
                                         fa_param=""))
                                curr_reac_id += 1
                                break

    return edge_list



import pandas as pd
import numpy as np
from typing import List, Dict
from .lipid import Lipid, FA
import itertools
import warnings
from .exceptions import ReferenceLipidError


class ReferenceLipid:

    __slots__ = ["_name", "_abbreviation", "_headgroup", "_potentialFAs", "_FAs", "_ether", "_hasLCB", "_altabb",
                 "_backbone", "_category"]

    def __init__(self, name: str, abbreviation: str, hg: str, potentialfas: int, fas: int, ether: int, haslcb: int,
                 altabb: str = None, backbone: str = "glycerol", category: str = "Glycerolipids"):
        self._name = name
        self._abbreviation = abbreviation
        self._headgroup = hg
        self._potentialFAs = potentialfas
        self._FAs = fas
        self._ether = ether
        self._hasLCB = haslcb
        self._altabb = altabb
        self._backbone = backbone
        self._category = category

        if self._FAs < (self._ether+self._hasLCB):
            raise ArithmeticError("The total number of FAs must be bigger/equal that the sum of ether + LCB FAs")

    def __str__(self) -> str:
        tmp = f"Reference class of {self._name}\n" \
              f"====\n" \
              f"* Abbreviation: {self._abbreviation}\n" \
              f"* Headgroup: {self._headgroup}\n" \
              f"* potential FAs: {self._potentialFAs}\n" \
              f"* FAs: {self._FAs}\n" \
              f"* Ehter bonds: {self._ether}\n" \
              f"* Has LCB: {self._hasLCB}\n" \
              f"* Alternative Abbreviation: {self._altabb}"
        return tmp

    def __repr__(self) -> str:
        return f"Reference class of {self._name}"

    def is_of_class(self, lipid: Lipid) -> bool:
        return (lipid.get_number_fas() == self._FAs) and (lipid.get_headgroup().lower() == self._headgroup.lower()) \
                and (lipid.get_lipid_class().lower() == self._abbreviation.lower()) and \
                (lipid.get_number_long_chain_base() == self._hasLCB) and (self._ether == lipid.get_ether())

    def get_name(self) -> str:
        return self._name

    def get_abbr(self) -> str:
        return self._abbreviation

    def get_headgroup(self) -> str:
        return self._headgroup

    def get_potentialfas(self) -> int:
        return self._potentialFAs

    def get_fas(self) -> int:
        return self._FAs

    def get_ether(self) -> int:
        return self._ether

    def get_lcb(self) -> int:
        return self._hasLCB

    def get_all_abbr(self) -> list:
        if (self._altabb is None) or (self._altabb == ""):
            return [self._abbreviation]
        else:
            return [self._abbreviation] + self._altabb.split("/")

    def get_backbone(self) -> str:
        return self._backbone

    def get_category(self) -> str:
        return self._category


def check_reference(lipid: Lipid, reference: ReferenceLipid, throw_error: bool = False) -> Lipid:
    if not reference.is_of_class(lipid):
        if throw_error:
            raise ReferenceLipidError(f"{str(lipid)} does not match with Reference class ({repr(reference)})")
        else:
            warnings.warn(f"{str(lipid)} does not match with Reference class ({repr(reference)})")

    return lipid


def parse_lipid_reference_table(tab: pd.DataFrame) -> List[ReferenceLipid]:

    return_l = []

    for row in range(tab.shape[0]):
        if bool(tab["Lipid Class"][row]):

            if tab["Lipid class"][row] is np.nan:
                raise KeyError("Each lipid class needs a name")
            else:
                dummy_name = str(tab["Lipid class"][row])

            if tab["Abbreviation"][row] is np.nan:
                raise KeyError("Each lipid class needs an abbreviation")
            else:
                dummy_abbr = str(tab["Abbreviation"][row])

            if tab["AltAbb"][row] is np.nan:
                dummy_altabb = ""
            else:
                dummy_altabb = str(tab["AltAbb"][row])

            if tab["HeadGroup"][row] is np.nan:
                dummy_hg = ""
            else:
                dummy_hg = str(tab["HeadGroup"][row]).upper()

            if tab["PotentialFAs"][row] is np.nan:
                dummy_pFA = 0
            else:
                dummy_pFA = int(tab["PotentialFAs"][row])

            if tab["FAs"][row] is np.nan:
                dummy_fas = 0
            else:
                dummy_fas = int(tab["FAs"][row])

            if tab["Ether"][row] is np.nan:
                dummy_ether = 0
            else:
                dummy_ether = int(tab["Ether"][row])

            if tab["has_long_chain_base"][row] is np.nan:
                dummy_lcb = 0
            else:
                dummy_lcb = int(tab["has_long_chain_base"][row])

            if tab["backbone"][row] is np.nan:
                dummy_backbone = "glycerol"
            else:
                dummy_backbone = str(tab["backbone"][row])

            if tab["Lipid category"][row] is np.nan:
                dummy_category = ""
            else:
                dummy_category = str(tab["Lipid category"][row])

            return_l.append(ReferenceLipid(name=dummy_name,
                                           abbreviation=dummy_abbr,
                                           hg=dummy_hg,
                                           potentialfas=dummy_pFA,
                                           fas=dummy_fas,
                                           ether=dummy_ether,
                                           haslcb=dummy_lcb,
                                           altabb=dummy_altabb,
                                           backbone=dummy_backbone,
                                           category=dummy_category))

    return return_l


def parse_lipid_reference_table_dict(tab: pd.DataFrame) -> Dict[str, ReferenceLipid]:
    tmp_list = parse_lipid_reference_table(tab)
    out_dict = {}
    for rl in tmp_list:
       for abb in rl.get_all_abbr():
           if abb in out_dict.keys():
               raise KeyError(f"Abbreviation '{abb}' already in use for another lipid.")
           out_dict[abb.upper()] = rl
    return out_dict


# def generate_reference_lipids_list(rl: List[ReferenceLipid], fa: List[FA], lcb: List[FA],
#                                    ether: List[FA]) -> List[Lipid]:
#     return_l = []
#
#     for f in fa:
#         if f.is_long_chain_base():
#             raise KeyError("Fatty Acids cannot be a long chain base")
#         if f.is_ether_bond():
#             raise KeyError("Fatty acid cannot be an ether bond")
#
#     for l in lcb:
#         if not l.is_long_chain_base():
#             raise KeyError("LCB has to be a LCB")
#         if l.is_ether_bond():
#             raise KeyError("LCB cannot be an ether bond")
#
#     for e in ether:
#         if not e.is_ether_bond():
#             raise KeyError("Ether FAs have to be ether")
#         if e.is_long_chain_base():
#             raise KeyError("Ether FA cannot be a LCB")
#
#     for lipid in rl:
#         n_fa = lipid.get_fas()
#         n_lcb = lipid.get_lcb()
#         n_ether = lipid.get_ether()
#
#         if (n_lcb == 0) and (n_ether == 0):
#             if len(fa) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one FA.")
#             combs = list(itertools.combinations_with_replacement(fa, n_fa))
#             for i in combs:
#                 return_l.append(Lipid(lipid_class=lipid.get_abbr(),
#                                       head_group=lipid.get_headgroup(),
#                                       is_molecular_species=True,
#                                       fa_spots=lipid.get_potentialfas(),
#                                       fatty_acids=list(i)))
#         elif (n_lcb == 0) and (n_ether != 0):
#             if len(fa) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one FA.")
#             if len(ether) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one ether FA.")
#
#             combs_fa = list(itertools.combinations_with_replacement(fa, n_fa-n_ether))
#             combs_ether = list(itertools.combinations_with_replacement(ether, n_ether))
#
#             for cf in combs_fa:
#                 for ce in combs_ether:
#                     return_l.append(Lipid(lipid_class=lipid.get_abbr(),
#                                           head_group=lipid.get_headgroup(),
#                                           is_molecular_species=True,
#                                           fa_spots=lipid.get_potentialfas(),
#                                           fatty_acids=list(cf)+list(ce)))
#
#         elif (n_lcb != 0) and (n_ether == 0):
#             if len(fa) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one FA.")
#             if len(lcb) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one LCB.")
#
#             combs_fa = list(itertools.combinations_with_replacement(fa, n_fa-n_lcb))
#             combs_lcb = list(itertools.combinations_with_replacement(lcb, n_lcb))
#
#             for cf in combs_fa:
#                 for cl in combs_lcb:
#                     return_l.append(Lipid(lipid_class=lipid.get_abbr(),
#                                           head_group=lipid.get_headgroup(),
#                                           is_molecular_species=True,
#                                           fa_spots=lipid.get_potentialfas(),
#                                           fatty_acids=list(cf)+list(cl)))
#         else:
#             if len(fa) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one FA.")
#             if len(lcb) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one LCB.")
#             if len(ether) == 0:
#                 raise ArithmeticError(f"Lipid {lipid.get_abbr()} requires at least one ether FA.")
#
#             combs_fa = list(itertools.combinations_with_replacement(fa, n_fa - n_lcb))
#             combs_lcb = list(itertools.combinations_with_replacement(lcb, n_lcb))
#             combs_ether = list(itertools.combinations_with_replacement(ether, n_ether))
#
#             for cf in combs_fa:
#                 for cl in combs_lcb:
#                     for ce in combs_ether:
#                         return_l.append(Lipid(lipid_class=lipid.get_abbr(),
#                                               head_group=lipid.get_headgroup(),
#                                               is_molecular_species=True,
#                                               fa_spots=lipid.get_potentialfas(),
#                                               fatty_acids=list(cf) + list(cl)+list(ce)))
#
#     return return_l
#
#
# def generate_reference_lipids_dict(rl: List[ReferenceLipid], fa: List[FA], lcb: List[FA],
#                                    ether: List[FA]) -> Dict[str, List[Lipid]]:
#
#     tmp_l = generate_reference_lipids_list(rl, fa, lcb, ether)
#
#     out_dict = {}
#
#     for i in tmp_l:
#         if i.get_lipid_class() in out_dict.keys():
#             out_dict[i.get_lipid_class()].append(i)
#         else:
#             out_dict[i.get_lipid_class()] = [i]
#
#     return out_dict



L_HGmodify = "L_HGmodify"
L_HGdelete = "L_HGdelete"
L_FAdelete = "L_FAdelete"
L_FAmodify = "L_FAmodify"
L_FAether = "L_FAether"
L_merge = "L_merge"
LCB_generation = "LCB_generation"
FFA_modify = "FFA_modify"
FFA_elong = "FFA_elong"
FFA_oxi = "FFA_oxi"
FFA_sat = "FFA_sat"
FFA_desat = "FFA_desat"
FFA_delete = "FFA_delete"



from typing import Dict, Tuple, List, Union
import networkx as nx
from networkx.algorithms import bipartite
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import math
import itertools
import re

# From package
from .reaction_types import L_FAmodify
from .edgelist import enzyme_str
from .libAP_nx import LSOprimizer, flatten, convergence_plot
from .utils import plot_networks, compute_ratios, plot_lineplot, plot_histogram, plot_boxplot


def analysis_reaction_network_enrichment(self,
                                         group1: str,
                                         group2: str,
                                         difference_as_change: bool = True,
                                         penalty_fa_reaction: float = 1,
                                         penalty_n_reactions: float = .2,
                                         default_ratio: float = 0,
                                         min_size: int = 5,
                                         max_size: int = 20,
                                         max_iter: int = 100,
                                         temp: float = 20,
                                         repetitions: int = 5,
                                         opt_max: bool = True,
                                         recompute_edgelist: bool = False,
                                         p_value_iterations: int = 10000,
                                         plot_scores: bool = True,
                                         plot_n_col: int = 3,
                                         plot_fig_size: Tuple[int, int] = (10, 18),
                                         plot_hspace: float = 0.4,
                                         correct_n_lipids: bool = True
                                         ) -> Tuple[Dict[str, nx.Graph],
                                                    Dict[str, List[float]],
                                                    Dict[str, float],
                                                    plt.Figure]:
    if not hasattr(self, "groups") or self.groups is None:
        raise AttributeError("Analysis only possible with group labels.")

    if (group2 not in self.groups.values) or (group1 not in self.groups.values):
        raise ValueError("Both given groups must be in the groups.")

    if group1 == group2:
        raise ValueError("Both groups have the same group name.")

    g, l_mapping = self.gln.native_network(filter_duplicates=False,
                                           excluded_reaction_types=None,
                                           return_lipid_mapping=True,
                                           bipartite=False,
                                           # bipartite_type='hyper',
                                           recompute_edgelist=recompute_edgelist,
                                           multi=True  # Necessary for correct hypergraph conversion
                                           )

    # Get data & lipid to data mapping
    user_data = self.data.T
    reverse_mapping = {
        net_name: data_name for data_name, net_names in l_mapping.items()
        for net_name in net_names
    }

    # ##############
    # ANALYSIS START------------------------------------------------------------------------------------------------
    # ##############
    print("Analysis start")

    # 1. Get all reaction IDs & edges with attributes---------------------------------------------------------------
    reaction_ids: Dict[str, List[List[Union[str, Dict]]]] = {}
    for ed in g.edges:
        tmp_edge_data = g.get_edge_data(*ed)
        if tmp_edge_data["reaction_id"] in reaction_ids.keys():
            reaction_ids[tmp_edge_data["reaction_id"]].append([ed[0], ed[1], tmp_edge_data])
        else:
            reaction_ids[tmp_edge_data["reaction_id"]] = [[ed[0], ed[1], tmp_edge_data]]

    print("Computing objectives")
    # 2. Compute ratio differences----------------------------------------------------------------------------------
    new_reaction_node_dict: Dict[str, List[Union[List[str], float]]] = {}

    for r_id, el in reaction_ids.items():
        # 2.1. Only one Edge -> Ratio or Penalty
        if len(el) == 1:
            nel = el[0]
            # 2.1.1. Penalty
            if nel[2]['reaction_type'] == L_FAmodify:
                new_reaction_node_dict[r_id] = [[nel[0], nel[1]], -penalty_fa_reaction, [nel[0], nel[1]]]
            # 2.1.2. Ratio
            else:
                tmp_ed = [reverse_mapping[nel[0]], reverse_mapping[nel[1]]]
                if difference_as_change:
                    tmp_ratio = user_data[tmp_ed[0]] - user_data[tmp_ed[1]]
                else:
                    tmp_ratio = user_data[tmp_ed[0]] / user_data[tmp_ed[1]]

                # Join with group labels
                tmp_ratio_df = pd.DataFrame({"r": tmp_ratio}).join(pd.DataFrame({"groups": self.groups}))

                # Series with groups
                tmp_series = tmp_ratio_df.groupby("groups").mean(numeric_only=True)["r"]
                if tmp_series[group1] == 0:
                    val = 0
                else:
                    val = abs((tmp_series[group1] - tmp_series[group2]) / tmp_series[group1])
                if val is np.nan or pd.isna(val):
                    val = default_ratio
                new_reaction_node_dict[r_id] = [[nel[0], nel[1], enzyme_str + nel[2]["enzyme_id"]], val,
                                                [nel[0], nel[1]]]
        # 2.2. More edges -> 1,2 or 2,2 reaction where partners need to be figured out
        elif len(el) == 2:
            nel0 = el[0]
            nel1 = el[1]
            substrates = []
            products = []
            if nel0[2]['l1_type'] == 'substrate':
                substrates.append(nel0[0])
                products.append(nel0[1])
            else:
                substrates.append(nel0[1])
                products.append(nel0[0])
            if nel1[2]['l1_type'] == 'substrate':
                substrates.append(nel1[0])
                products.append(nel1[1])
            else:
                substrates.append(nel1[1])
                products.append(nel1[0])
            mapped_substrates = list(set([reverse_mapping[x] for x in substrates]))
            mapped_products = list(set([reverse_mapping[x] for x in products]))

            if difference_as_change:
                if correct_n_lipids:
                    tmp_ratio = np.mean([user_data[x] for x in mapped_products], axis=0) - np.mean(
                        [user_data[x] for x in mapped_substrates], axis=0)
                    tmp_ratio = pd.Series(tmp_ratio, index=user_data[mapped_products[0]].index)
                else:
                    tmp_ratio = sum([user_data[x] for x in mapped_products]) - sum(
                        [user_data[x] for x in mapped_substrates])
            else:
                if correct_n_lipids:
                    product_prod = math.prod([user_data[x] for x in mapped_substrates]) ** (1 / len(mapped_substrates))
                    substrate_prod = math.prod([user_data[x] for x in mapped_products]) ** (1 / len(mapped_products))
                    tmp_ratio = product_prod / substrate_prod
                else:
                    tmp_ratio = math.prod([user_data[x] for x in mapped_products]) / \
                                math.prod([user_data[x] for x in mapped_substrates])

            # Join with group labels
            tmp_ratio_df = pd.DataFrame({"r": tmp_ratio}).join(pd.DataFrame({"groups": self.groups}))

            # Series with groups
            tmp_series = tmp_ratio_df.groupby("groups").mean(numeric_only=True)["r"]
            if tmp_series[group1] == 0:
                val = 0
            else:
                val = abs((tmp_series[group1] - tmp_series[group2]) / tmp_series[group1])
            if val is np.nan or pd.isna(val):
                val = default_ratio
            ids = flatten([substrates, products, [enzyme_str + nel0[2]["enzyme_id"]]])
            new_reaction_node_dict[r_id] = [list(set(ids)), val, flatten([substrates, products])]

        else:
            raise ValueError("More edges than expected occurred per reaction. Please contact the developer")
    # 3. Build & reconnect network----------------------------------------------------------------------------------
    print("Building & reconnecting network")

    # Prepare nodes
    new_node_list = []
    node_set = dict()
    counter = 0
    for n, v in new_reaction_node_dict.items():
        if str(sorted(v[2])) not in node_set.keys():
            new_node_list.append((n, {'org_nodes': v[0], 'objective': v[1], 'enzymes': [x for x in v[0] if
                                                                                        x.startswith(enzyme_str)],
                                      'lipid_nodes': v[2]}))
            node_set[str(sorted(v[2]))] = counter
            counter += 1
        else:
            new_node_list[node_set[str(sorted(v[2]))]][1]['org_nodes'] += [x for x in v[0] if x.startswith(enzyme_str)]
            new_node_list[node_set[str(sorted(v[2]))]][1]['enzymes'] += [x for x in v[0] if x.startswith(enzyme_str)]

    # Make bipartite graph
    B = nx.Graph()
    B.add_nodes_from(new_node_list, bipartite=0)
    B.add_nodes_from(list(g.nodes), bipartite=1)  # Lipids
    enz_l = []
    for i in new_node_list:
        if len(i[1]['enzymes']) > 0:
            enz_l += i[1]['enzymes']
    B.add_nodes_from(list(set(enz_l)), bipartite=1)  # Enzymes

    new_edgelist = []
    for i in new_node_list:
        for j in i[1]['org_nodes']:
            new_edgelist.append((i[0], j))
    B.add_edges_from(new_edgelist)  # Edges between reactions and Lipids/Enzymes

    # Project back to reaction network
    top_nodes = {n for n, d in B.nodes(data=True) if d["bipartite"] == 0}
    # bottom_nodes, _ = bipartite.sets(B)
    G = bipartite.projected_graph(B, top_nodes)

    # # Make edges
    # new_edgelist = set()
    # for nn1 in range(len(new_node_list)):
    #     for nn2 in range(nn1, len(new_node_list)):
    #         if nn1 != nn2:
    #             # share lipids or reactions make an edge
    #             if len(set(new_node_list[nn1][1]['org_nodes']).intersection(new_node_list[nn2][1]['org_nodes'])) > 0:
    #                 new_edgelist.add((new_node_list[nn1][0], new_node_list[nn2][0]))

    gn = nx.Graph()                  # Reaction graph
    gn.add_nodes_from(new_node_list)
    gn.add_edges_from(list(G.edges))
    # gn.add_edges_from(new_edgelist)

    # 4. Local search-----------------------------------------------------------------------------------------------
    print("Local search")
    gh, l_mapping = self.gln.native_network(filter_duplicates=False,
                                            excluded_reaction_types=None,
                                            return_lipid_mapping=True,
                                            bipartite=True,
                                            bipartite_type='hyper',
                                            recompute_edgelist=recompute_edgelist
                                            )
    results = {}
    scores = {}
    p_values = {}
    for c in nx.connected_components(gn):
        if len(c) > max_size:
            cg = gn.subgraph(c).copy()
            sc = -100000
            nodes = []
            c_scores = []
            for rep in range(repetitions):
                print("Local search repetition:", rep)
                opt = LSOprimizer(G=cg,
                                  L_min=min_size,
                                  L_max=max_size,
                                  max_iter=max_iter,
                                  reaction_penalty=penalty_n_reactions,
                                  T=temp,
                                  verbose=False,
                                  node_scoring=True,
                                  plot=False)
                tmp_nodes, tmp_sc, tmp_c_scores = opt.run_ls()
                if tmp_sc > sc:
                    print(f"Improved to {tmp_sc}")
                    sc = tmp_sc
                    nodes = tmp_nodes
                    c_scores = tmp_c_scores

            if opt_max:
                print("Rerunning local search on best result")
                opt = LSOprimizer(G=cg,
                                  L_min=min_size,
                                  L_max=max_size,
                                  max_iter=max_iter,
                                  reaction_penalty=penalty_n_reactions,
                                  T=temp,
                                  verbose=False,
                                  node_scoring=True,
                                  seed=nodes,
                                  plot=False)
                tmp_nodes, tmp_sc, tmp_c_scores = opt.run_ls()
                if tmp_sc > sc:
                    print(f"Improved to {tmp_sc}")
                    sc = tmp_sc
                    nodes = tmp_nodes
                    c_scores = c_scores + tmp_c_scores
            if plot_scores:
                convergence_plot(c_scores)
            # Transform back to original graph
            tmp_na = dict(gn.nodes(data='org_nodes'))
            transformed_nodes = list(set(flatten([tmp_na[x] for x in nodes])))
            g_small = nx.subgraph(gh, transformed_nodes)

            title = f'Connected component {len(results)}'
            results[title] = g_small
            scores[title] = c_scores

            # 5. Compute empirical p-value for result ------------------------------------------------------------------
            print("Computing p-value")
            tmp_seed_score = []
            for p_iter in range(p_value_iterations):
                n_sample_nodes = np.random.choice(np.arange(min_size, max_size+1), 1)[0]
                tmp_seed_nodes = np.random.choice(list(new_reaction_node_dict.keys()), n_sample_nodes, replace=False)
                tmp_score = [new_reaction_node_dict[key][1] for key in tmp_seed_nodes if
                                                new_reaction_node_dict[key][1] != -penalty_fa_reaction]
                if len(tmp_score) > 0:
                    tmp_seed_score.append(np.mean(tmp_score))
                else:
                    tmp_seed_score.append(0)
            p_values[title] = float(np.sum(np.array(tmp_seed_score) > sc)) / p_value_iterations

    results, fig = plot_networks(
        networks=results,
        scores=scores,
        p_values=p_values,
        plot_n_col=plot_n_col,
        plot_fig_size=plot_fig_size,
        plot_hspace=plot_hspace
    )

    return results, scores, p_values, fig


def analysis_ratio_network_enrichment(self,
                                      group1: str,
                                      group2: str,
                                      data_is_log: bool = True,
                                      penalty_fa_reaction: float = 1,
                                      penalty_n_reactions: float = 1,
                                      default_ratio: float = 0,
                                      min_size: int = 5,
                                      max_size: int = 20,
                                      max_iter: int = 100,
                                      temp: float = 20,
                                      recompute_edgelist: bool = False,
                                      hyper: bool = False) -> List[nx.Graph]:

    if not hasattr(self, "groups") or self.groups is None:
        raise AttributeError("Analysis only possible with group labels.")

    if (group2 not in self.groups.values) or (group1 not in self.groups.values):
        raise ValueError("Both given groups must be in the groups.")

    if hyper:
        g, l_mapping = self.gln.native_network(filter_duplicates=False,
                                               excluded_reaction_types=None,
                                               return_lipid_mapping=True,
                                               bipartite=True,
                                               bipartite_type='hyper',
                                               recompute_edgelist=recompute_edgelist
                                               )
    else:
        g, l_mapping = self.gln.native_network(filter_duplicates=False,
                                               excluded_reaction_types=None,
                                               return_lipid_mapping=True,
                                               bipartite=False,
                                               recompute_edgelist=recompute_edgelist
                                               )

    user_data = self.data.T
    reverse_mapping = {
        net_name: data_name for data_name, net_names in l_mapping.items()
        for net_name in net_names
    }

    # Compute ratios
    for ed in g.edges:
        tmp_edge_data = g.get_edge_data(*ed)
        # print(ed[1])
        if (tmp_edge_data['reaction_type'] != L_FAmodify) and ((not ed[1].startswith(enzyme_str)) and
                                                               (not ed[0].startswith(enzyme_str))):
            tmp_ed = [reverse_mapping[ed[0]], reverse_mapping[ed[1]]]
            if data_is_log:
                tmp_ratio = user_data[tmp_ed[0]] - user_data[tmp_ed[1]]
            else:
                tmp_ratio = user_data[tmp_ed[0]] / user_data[tmp_ed[1]]

            # Join with group labels
            tmp_ratio_df = pd.DataFrame({"r": tmp_ratio}).join(pd.DataFrame({"groups": self.groups}))

            # Series with groups
            tmp_series = tmp_ratio_df.groupby("groups").mean(numeric_only=True)["r"]
            val = abs((tmp_series[group1] - tmp_series[group2]) / tmp_series[group1])
            if val is np.nan or pd.isna(val):
                val = default_ratio
            g.edges[ed[0], ed[1]]["objective"] = val

        elif tmp_edge_data['reaction_type'] == L_FAmodify and ((not ed[1].startswith(enzyme_str)) and
                                                               (not ed[0].startswith(enzyme_str))):
            # Reaction which will be penalized
            g.edges[ed[0], ed[1]]["objective"] = - penalty_fa_reaction

        else:
            g.edges[ed[0], ed[1]]["objective"] = 0

    if hyper:  # Local search not yet implemented for hyper
        return [g]
    # Local search
    results = []
    for c in nx.connected_components(g):
        if len(c) > max_size:
            cg = g.subgraph(c).copy()
            opt = LSOprimizer(G=cg,
                              L_min=min_size,
                              L_max=max_size,
                              max_iter=max_iter,
                              reaction_penalty=penalty_n_reactions,
                              T=temp,
                              verbose=False)
            nodes, sc = opt.run_ls()
            g_small = nx.subgraph(g, nodes)

            pos = nx.spring_layout(g_small)
            labels = nx.get_edge_attributes(g_small, 'objective')
            for key in labels:
                labels[key] = np.round(labels[key], 2)
            nx.draw_networkx(g_small, pos, font_size=8)

            nx.draw_networkx_edge_labels(g_small, pos, edge_labels=labels)
            plt.title(f'connected component {len(results)}, score {np.round(sc, 3)}')
            plt.show()

            results.append(g_small)

    return results


def _extract_reactions_(
    network: nx.Graph, edge_attr: str
) -> Tuple[List[str], Dict[str, str]]:
    unique_enzyme_list = []
    enzyme_structure_dict = {}
    for ed in network.edges:
        tmp_edge_data = network.get_edge_data(*ed)
        if tmp_edge_data['reaction_type'] != L_FAmodify:
            if tmp_edge_data[edge_attr] not in unique_enzyme_list:
                unique_enzyme_list.append(tmp_edge_data[edge_attr])
                if tmp_edge_data["reaction_structure"] == "1,1":
                    enzyme_structure_dict[tmp_edge_data[edge_attr]] = tmp_edge_data["reaction_structure"]
                else:
                    enzyme_structure_dict[tmp_edge_data[edge_attr]] = tmp_edge_data["reaction_structure"] + \
                                                                      ";" + tmp_edge_data['reaction_type']
    return unique_enzyme_list, enzyme_structure_dict


def _ratio_distributions_(
    self,
    network: nx.Graph,
    name_map: Dict[str, List[str]],
    enzyme_subset: Union[None, List] = None,
    difference_as_change: bool = True,
    distribution_per_sample: bool = True,
    select_groups: List[str] = None,
    fa_specific: bool = False,
    plot_n_col: int = 3,
    plot_fig_size: Tuple[int, int] = (20, 20),
    plot_hspace: float = 0.4,
    difference_dist: bool = False,
    reference_group: str = "",
    as_boxplot: bool = False,
    as_subplots: bool = True,
    z_scores: bool = False
) -> Tuple[plt.Figure, Dict[str, pd.DataFrame]]:
    # Extract all reactions classes for reaction_type != L_FAmodify (only class reactions)
    if fa_specific:
        edge_attr = "id_fa_raw"
    else:
        edge_attr = "enzyme_id"

    unique_enzyme_list, enzyme_structure_dict = _extract_reactions_(network, edge_attr)

    print(f"Identified {len(unique_enzyme_list)} reaction classes for which ratio distribution analysis "
          f"will be performed.")

    if len(unique_enzyme_list) == 0:
        raise ValueError("No class reactions could be identified, cannot perform ratio distribution analysis.")

    user_data = self.data.T
    reverse_mapping = {
        net_name: data_name for data_name, net_names in name_map.items()
        for net_name in net_names
    }
    # removing enzymes if enzyme subset is given
    if enzyme_subset is not None:
        enzyme_subset = {enzyme.split('Enzyme: ')[1] for enzyme in enzyme_subset}
        unique_enzyme_list = list(set(unique_enzyme_list).intersection(enzyme_subset))
        enzyme_structure_dict = {enzyme: enzyme_structure_dict[enzyme] for enzyme in unique_enzyme_list}

    # Compute ratios for all reaction classes -> per sample and per reaction
    # TODO:
    # * 2,2 & 1,2 reactions using reaction_structure and l1/l2_type attributes
    # * for 2,2 L_HG the set_id has to be considered to make two sets
    df_dict = compute_ratios(unique_enzyme_list=unique_enzyme_list,
                             enzyme_structure_dict=enzyme_structure_dict,
                             gr=network,
                             reverse_mapping=reverse_mapping,
                             difference_as_change=difference_as_change,
                             user_data=user_data,
                             class_reactions=self.gln.get_class_reactions(),
                             edge_attr=edge_attr)

    # Prepare data for histogram
    # If groups are given, then add those to the df
    # The ratios should be in the column "value" and potential groups in column "groups"
    if hasattr(self, "groups") and self.groups is not None:
        if distribution_per_sample:
            data_for_hist = {key: df.melt(ignore_index=False).join(
                pd.DataFrame({"groups": self.groups})).reset_index() for key, df in df_dict.items()}

            if select_groups is not None:
                data_for_hist = {key: df[df["groups"].isin(select_groups)] for key, df in data_for_hist.items()}
        else:
            data_for_hist = {key: df.join(
                pd.DataFrame({"groups": self.groups})).groupby(
                "groups").mean(numeric_only=True).melt(
                ignore_index=False).reset_index() for key, df in df_dict.items()}

            if difference_dist:
                tmp_dict = dict()
                for reac, dat in data_for_hist.items():
                    ref_dat = dat[dat['groups'] == reference_group].set_index('variable')
                    oth_dat = dat[dat['groups'] != reference_group].set_index('variable')
                    tmp_dat = ref_dat.join(oth_dat, how='left', lsuffix="_l", rsuffix="_r")
                    tmp_dat = tmp_dat.assign(**{
                        "value": (tmp_dat['value_r'] - tmp_dat['value_l']) / abs(tmp_dat['value_l'])
                    }).reset_index().rename(columns={"groups_r": "groups"})
                    tmp_dict[reac] = tmp_dat
                data_for_hist = tmp_dict

            if select_groups is not None:
                data_for_hist = {key: df[df["groups"].isin(select_groups)] for key, df in data_for_hist.items()}
    else:
        if distribution_per_sample:
            data_for_hist = {key: df.melt(ignore_index=False) for key, df in df_dict.items()}
        else:
            data_for_hist = {key: pd.DataFrame({"value": df.mean(numeric_only=True)}) for key, df in
                             df_dict.items()}
    if as_boxplot:
        fig = plot_boxplot(data_for_hist,
                           plot_n_col=plot_n_col,
                           plot_fig_size=plot_fig_size,
                           plot_hspace=plot_hspace,
                           difference_dist=difference_dist,
                           as_subplots=as_subplots,
                           z_scores=z_scores)
    else:
        fig = plot_histogram(data_for_hist,
                             plot_n_col=plot_n_col,
                             plot_fig_size=plot_fig_size,
                             plot_hspace=plot_hspace,
                             difference_dist=difference_dist)

    # TODO: 4. Distribution deconvolution

    return fig, data_for_hist


def analysis_ratio_distribution(self,
                                enzyme_subset: List = None,
                                difference_as_change: bool = True,
                                distribution_per_sample: bool = True,
                                select_groups: List[str] = None,
                                fa_specific: bool = False,
                                plot_n_col: int = 3,
                                plot_fig_size: Tuple[int, int] = (20, 20),
                                plot_hspace: float = 0.4,
                                difference_dist: bool = False,
                                reference_group: str = "",
                                as_boxplot: bool = False,
                                as_subplots: bool = True,
                                z_scores: bool = False) -> Tuple[plt.Figure, Dict[str, pd.DataFrame]]:
    """
    TODO: Tim
    :param enzyme_subset: list, default None, if given only reactions with the given enzymes are plotted
    :param difference_as_change: boolean, default True, to compute the difference or the quotient between substrates and products
    :param distribution_per_sample: boolean, default True, whether to plot distribution for each sample or cumulated
    :param select_groups: list, default None, list of groups to consider in the analysis
    :param fa_specific: boolean, default False, whether to compute ratios fatty acid specific
    :param plot_n_col: int, default 3, number of subplots per row
    :param plot_fig_size: Tuple[int, int], default (20, 20), total figure size
    :param plot_hspace: float, default 0.4, horizontal spacing between subplots
    :param difference_dist: boolean, default False, whether to plot the difference of group distributions or the
                            distributions per group
    :param reference_group: str, default '',
    :param as_boxplot: boolean, default False, whether to plot ratios as boxplots or histograms (if False)
    :param as_subplots: boolean default True, if False and as_boxplot is True all reactions will be plotted in a
                        single plot with one reaction per position on the x-axis
    :param z_scores: boolean, default False, whether to scale ratios to zero mean and unit variance (over all groups)
    :return: 2-tuple, plt.Figure containing histograms and a dictionary of data frames containing ratio distributions
    """
    if (not hasattr(self, "groups") or self.groups is None) and difference_dist:
        raise AttributeError("Analysis with difference_dist=True only possible with group labels.")

    if difference_dist and (reference_group not in self.groups.values):
        raise ValueError("Reference group must be in the groups.")
    # Recompute network (native, bipartite=False)
    print("Recomputing native network (This will not overwrite your last computed network).")
    g, l_mapping = self.gln.native_network(filter_duplicates=False,
                                           excluded_reaction_types=None,
                                           return_lipid_mapping=True,
                                           bipartite=False,
                                           multi=True
                                           )
    return _ratio_distributions_(
        self, g, l_mapping, enzyme_subset,
        difference_as_change, distribution_per_sample, select_groups,
        fa_specific, plot_n_col, plot_fig_size, plot_hspace,
        difference_dist, reference_group, as_boxplot, as_subplots,
        z_scores=z_scores
    )


def analysis_enrichment_ratio_plots(
    self,
    enrichment_network: nx.Graph,
    difference_as_change: bool = True,
    distribution_per_sample: bool = True,
    select_groups: List[str] = None,
    fa_specific: bool = False,
    plot_n_col: int = 3,
    plot_fig_size: Tuple[int, int] = (20, 20),
    plot_hspace: float = 0.4,
    difference_dist: bool = False,
    reference_group: str = "",
    as_boxplot: bool = False,
    as_subplots: bool = True,
    z_scores: bool = True
) -> Tuple[plt.Figure, Dict[str, pd.DataFrame]]:
    """
    :param enrichment_network: nx.Graph, enriched subnetwork as returned by
                               `LipidNetwork.analysis_reaction_network_enrichment`
    :param difference_as_change: boolean, default True, to compute the difference or the quotient between substrates and products
    :param distribution_per_sample: boolean, default True, whether to plot distribution for each sample or cumulated
    :param select_groups: list, default None, list of groups to consider in the analysis
    :param fa_specific: boolean, default False, whether to compute ratios fatty acid specific
    :param plot_n_col: int, default 3, number of subplots per row
    :param plot_fig_size: Tuple[int, int], default (20, 20), total figure size
    :param plot_hspace: float, default 0.4, horizontal spacing between subplots
    :param difference_dist: boolean, default False, whether to plot the difference of group distributions or the
                            distributions per group
    :param reference_group: str, default '',
    :param as_boxplot: boolean, default False, whether to plot ratios as boxplots or histograms (if False)
    :param as_subplots: boolean default True, if False and as_boxplot is True all reactions will be plotted in a
                        single plot with one reaction per position on the x-axis
    :param z_scores: boolean, default False, whether to scale ratios to zero mean and unit variance (over all groups)
    :return: 2-tuple, plt.Figure containing histograms and a dictionary of data frames containing ratio distributions
    """
    if (not hasattr(self, "groups") or self.groups is None) and difference_dist:
        raise AttributeError("Analysis with difference_dist=True only possible with group labels.")
    if difference_dist and (reference_group not in self.groups.values):
        raise ValueError("Reference group must be in the groups.")
    print("Computing native network. This will not overwrite your previously computed network")
    native_net, lipid_mapping = self.gln.native_network(
        filter_duplicates=False,
        excluded_reaction_types=None,
        return_lipid_mapping=True,
        bipartite=False
    )

    direct_edges = [edge for edge in enrichment_network.edges
                    if edge in native_net.edges and
                    f"Enzyme: {native_net.edges[edge]['enzyme_id']}" in enrichment_network.nodes]
    cleaned_network = enrichment_network.edge_subgraph(direct_edges)
    return _ratio_distributions_(
        self, cleaned_network, lipid_mapping, None,
        difference_as_change, distribution_per_sample, select_groups,
        fa_specific, plot_n_col, plot_fig_size, plot_hspace,
        difference_dist, reference_group, as_boxplot, as_subplots,
        z_scores=z_scores
    )


def analysis_chain_length(self,
                          control_condition: str,
                          other_groups: List[str] = None,
                          lipid_classes: List[str] = None,
                          data_is_log: bool = False,
                          return_data: bool = False,
                          **kwargs
                          ) -> Union[plt.Figure, Dict[str, pd.DataFrame]]:
    """
    TODO
    :param self:
    :param control_condition:
    :param other_groups:
    :param lipid_classes:
    :param data_is_log:
    :param return_data:
    :param kwargs:
    :return:
    """
    if not hasattr(self, "groups"):
        raise AttributeError("Data groups must be available for fold change Analysis")

    # Set lipid weights
    lipid_dict = self._get_weighted_lipid_dict()
    if lipid_classes is None:
        lipid_classes = lipid_dict.keys()
    feature_matrix = self.data.T.drop(self.data.T.index.values)

    for cls in lipid_dict.keys():
        if cls in lipid_classes:
            for lip in lipid_dict[cls]:

                # Get class & length
                tmp_feature = lip.get_lipid_class() + ": " + str(lip.sum_length())

                # Add row, if feature not yet in:
                if tmp_feature not in feature_matrix.index.values:
                    feature_matrix = feature_matrix.append(pd.Series([0] * feature_matrix.shape[1],
                                                                     index=feature_matrix.columns,
                                                                     name=tmp_feature))
                # TODO: double check division factor
                feature_matrix.loc[tmp_feature, lip.get_dataname()] += (1 / lip.get_div_factor())

    new_df = feature_matrix.dot(self.data).T

    classes = list(set([x.split(": ")[0] for x in new_df.columns.values]))

    # Get sub df only for one class
    df_list = {}

    for cls in sorted(classes):
        df_list[cls] = new_df[[col for col in new_df if col.startswith(f'{cls}:')]].copy()
        df_list[cls] = df_list[cls].rename(columns={col: col.split(": ")[1] for col in df_list[cls].columns.values})
        joined_df = df_list[cls].join(pd.DataFrame({"groups": self.groups}))
        control_df = joined_df[joined_df["groups"] == control_condition]
        control_mean = pd.DataFrame({"val": control_df.mean(axis=0,
                                                            numeric_only=True)}).reset_index().rename(
            columns={"index": "variable"})

        if other_groups is not None:
            other_df = joined_df[joined_df["groups"].isin(other_groups)].groupby("groups").mean().melt(
                ignore_index=False).reset_index()
        else:
            other_df = joined_df[joined_df["groups"] != control_condition].groupby("groups").mean().melt(
                # old:  != control_condition
                ignore_index=False).reset_index()
        other_df = other_df.set_index('variable').join(control_mean.set_index('variable')).reset_index()
        if data_is_log:
            other_df["FC"] = other_df["value"] - other_df["val"]
        else:
            other_df["FC"] = other_df["value"] / other_df["val"]

        df_list[cls] = other_df

    if return_data:
        return df_list
    res = plot_lineplot(df_list, **kwargs)
    return res


def analysis_lipid_substructure(self, combined: bool = False) -> pd.DataFrame:
    """
    TODO: Tim

    Parameters
    ----------
    :param combined: bool, default=False
        Combine newly generated lipid substructures with the original lipid data and
        output the combined data frame.

    Returns
    -------
    :return: pd.DataFrame
        Data frame with lipid substructures as amounts for each sample.
        Optionally, the data frame is combined with the original lipid data.
    """
    # Set lipid weights
    lipid_dict = self._get_weighted_lipid_dict()

    # Construction of the feature weight matrix
    # This will later be multiplied with data matrix
    feature_matrix = self.data.T.drop(self.data.T.index.values)

    for cls in lipid_dict.keys():
        for lip in lipid_dict[cls]:
            # make list of all considered features
            tmp_features = list()
            # *category
            tmp_features.append(lip.get_category())
            # *backbone
            tmp_features.append(lip.get_backbone())
            # *headgroup
            tmp_features.append(lip.get_headgroup())
            # *class
            tmp_features.append(lip.get_lipid_class())
            # *fatty acids
            # **single
            for i in lip.get_fas():
                tmp_features.append(f"FA: {str(i)}")
            # **combined -> for now only pairwise
            if len(lip.get_fas()) >= 2:
                for x in itertools.combinations(lip.get_fas(), 2):
                    tmp_features.append(f"FA: {str(sorted(x))}")

            # 1. Single features
            for feat in tmp_features:
                # Add row, if feature not yet in:
                if feat not in feature_matrix.index.values:
                    feature_matrix = feature_matrix.append(pd.Series([0] * feature_matrix.shape[1],
                                                                     index=feature_matrix.columns,
                                                                     name=feat))

                feature_matrix.loc[feat, lip.get_dataname()] += (1 / lip.get_div_factor())

            # 2. Combined features
            for x in itertools.combinations(tmp_features, 2):
                combined_feature = str(sorted(x))
                if len(re.findall("FA", combined_feature)) < 2:
                    if combined_feature not in feature_matrix.index.values:
                        feature_matrix = feature_matrix.append(pd.Series([0] * feature_matrix.shape[1],
                                                                         index=feature_matrix.columns,
                                                                         name=combined_feature))

                    feature_matrix.loc[combined_feature, lip.get_dataname()] += (1 / lip.get_div_factor())
    if combined:
        return feature_matrix.dot(self.data).append(self.data)
    else:
        return feature_matrix.dot(self.data)




from __future__ import annotations
from typing import List
from .reference import ReferenceLipid
from .validation import reaction_validator
from .reaction_types import *
from .lipid import Lipid, FA


class Reaction:

    __slots__ = ["_substrates",
                 "_products",
                 "_Rtype",
                 "_enzymeIDs",
                 "_nl_participants",
                 "_uniprot",
                 "_gene_name"
                 ]

    def __init__(self,
                 substrates: List[ReferenceLipid],
                 products: List[ReferenceLipid],
                 reaction_type: str,
                 enzyme_ids: str,
                 non_lipid_reaction_participants: List[str],
                 uniprot: str = "",
                 gene_name: str = ""
                 ):
        self._substrates = substrates
        self._products = products
        self._Rtype = reaction_type
        self._enzymeIDs = enzyme_ids
        self._nl_participants = non_lipid_reaction_participants
        self._uniprot = uniprot
        if gene_name == "-":
            self._gene_name = ""
        self._gene_name = ", ".join(list(set(gene_name.upper().split(", "))))

    def is_valid(self, verbose=True) -> bool:
        return reaction_validator(self._Rtype, self._substrates, self._products, verbose=verbose)

    def shape(self) -> tuple:
        return len(self._substrates), len(self._products)

    def participation_equality(self, other: Reaction) -> bool:
        mysubs = sorted([x.get_abbr() for x in self._substrates])
        myprods = sorted([x.get_abbr() for x in self._products])
        osubs = sorted([x.get_abbr() for x in other._substrates])
        oprods = sorted([x.get_abbr() for x in other._products])

        return ((mysubs == osubs) and (myprods == oprods)) or ((mysubs == oprods) and (myprods == osubs))

    def subs_eq_prods(self) -> bool:
        mysubs = sorted([x.get_abbr() for x in self._substrates])
        myprods = sorted([x.get_abbr() for x in self._products])
        return mysubs == myprods

    def __eq__(self, other: Reaction) -> bool:
        return self.participation_equality(other) and self._Rtype == other._Rtype

    def extend_enzyme_ids(self, new_id: str):
        if new_id != "" and new_id is not None and new_id != "-":
            self._enzymeIDs += f";{new_id}"

    def extend_uniprot(self, new_id: str):
        if new_id != "" and new_id is not None and new_id != "-":
            if self._uniprot == "":
                self._uniprot = new_id
            else:
                self._uniprot += f", {new_id}"
            self._uniprot = ", ".join(list(set(self._uniprot.upper().split(", "))))

    def extend_gene_name(self, new_id: str):
        if new_id != "" and new_id is not None and new_id != "-":
            self._gene_name += f", {new_id}"
            self._gene_name = ", ".join(list(set(self._gene_name.upper().split(", "))))

    def extend_nl_participants(self, new_p: List[str]):
        self._nl_participants = list(set(self._nl_participants + new_p))

    def __str__(self) -> str:
        val = "Valid" if self.is_valid() else "Not-valid"
        return val + " reaction " + self._Rtype + ": " + str([x.get_abbr() for x in self._substrates])  + " <=> " + \
                     str([x.get_abbr() for x in self._products]) + " (" + self._enzymeIDs + ")"

    def reaction_nice_str(self) -> str:
        tmp = "Reaction " + self._Rtype + ":\n" + str([x.get_abbr() for x in self._substrates]) + " <=> " + \
               str([x.get_abbr() for x in self._products])
        if len(self.get_gene_name()) > 0:
            tmp += "\n(" + self.get_gene_name() + ")"
        tmp += "\n(" + self._enzymeIDs + ")"
        return tmp

    def get_reaction_type(self) -> str:
        return self._Rtype

    def get_enzyme_id(self) -> str:
        return self._enzymeIDs

    def get_enzyme_repr(self) -> str:
        if self._gene_name == "" or self._gene_name is None:
            return self._enzymeIDs
        else:
            return self._gene_name + "\n" + self._enzymeIDs

    def get_substrates(self) -> List[ReferenceLipid]:
        return self._substrates

    def get_products(self) -> List[ReferenceLipid]:
        return self._products

    def get_nl_participants(self) -> List[str]:
        return self._nl_participants

    def get_uniprot(self) -> str:
        return self._uniprot

    def get_gene_name(self) -> str:
        return self._gene_name

    def short_str(self, str_len=100, nice=False) -> str:
        if nice:
            tmp_str = self.reaction_nice_str()
        else:
            tmp_str = self.__str__()
        if len(tmp_str) > str_len:
            return tmp_str[0:str_len]+"..."
        else:
            return tmp_str


class FAReaction:

    __slots__ = ["__C", "__DB", "__OH", "__valid_ether", "__valid_lcb", "__last_conversion", "__reaction_name"]

    def __init__(self, c: int, db: int, oh: int, valid_for_ether: bool, valid_for_lcb, name: str = ""):
        self.__C = c
        self.__DB = db
        self.__OH = oh
        self.__valid_ether = valid_for_ether
        self.__valid_lcb = valid_for_lcb
        self.__last_conversion = ""
        self.__reaction_name = name

    def get_c(self) -> int:
        return self.__C

    def get_db(self) -> int:
        return self.__DB

    def get_oh(self) -> int:
        return self.__OH

    def valid_for_ether(self) -> bool:
        return self.__valid_ether

    def valid_for_lcb(self) -> bool:
        return self.__valid_lcb

    def get_name(self) -> str:
        return self.__reaction_name

    def is_transformable(self, l1: Lipid, l2: Lipid, excluded_reactions: List[List[FA]] = None) -> bool:

        if not l1.is_molecular_species() or not l2.is_molecular_species():
            raise ValueError("All lipids for fatty acid reactions must be molecular species.")

        # Lipids must be from the same class to be transformable
        if l1.get_lipid_class() != l2.get_lipid_class():
            return False
        if (((l1.sum_length() + self.__C) != l2.sum_length()) or ((l2.sum_length() + self.__C) != l1.sum_length())) and \
            (((l1.sum_dbs() + self.__DB) != l2.sum_dbs()) or ((l2.sum_dbs() + self.__DB) != l1.sum_dbs())) and \
            (((l1.sum_ohs() + self.__OH) != l2.sum_ohs()) or ((l2.sum_ohs() + self.__OH) != l1.sum_ohs())):
            return False

        fas1 = l1.get_fas()
        fas2 = l2.get_fas()
        for fa1 in range(len(fas1)):
            for fa2 in range(len(fas2)):

                # Check for FA types
                if fas1[fa1].is_long_chain_base() != fas2[fa2].is_long_chain_base():
                    continue
                if fas1[fa1].is_ether_bond() != fas2[fa2].is_ether_bond() or \
                        fas1[fa1].get_ether_type() != fas2[fa2].get_ether_type():
                    continue

                # Other FAs must be the same
                f1c = fas1.copy()
                f2c = fas2.copy()

                del f1c[fa1]
                del f2c[fa2]
                if sorted(f1c) != sorted(f2c):
                    continue

                # Check if FA types are valid here
                if (fas1[fa1].is_ether_bond() or fas2[fa2].is_ether_bond()) and not self.valid_for_ether():
                    continue
                if (fas1[fa1].is_long_chain_base() or fas2[fa2].is_long_chain_base()) and not self.valid_for_lcb():
                    continue

                excluded = False
                # Skip excluded reactions
                if excluded_reactions is not None:
                    for er in excluded_reactions:
                        if (fas1[fa1] == er[0] and fas2[fa2] == er[1]) or \
                                (fas1[fa1] == er[1] and fas2[fa2] == er[0]):
                            excluded = True
                            continue
                # Evaluate rule
                if not excluded:
                    if (fas1[fa1].modified_fa(self.__C, self.__DB, self.__OH) == fas2[fa2]) or \
                            (fas2[fa2].modified_fa(self.__C, self.__DB, self.__OH) == fas1[fa1]):
                        self.__last_conversion = f"{fas1[fa1]} - {fas2[fa2]}"
                        return True

        return False

    def get_last_conversion(self) -> str:
        return self.__last_conversion

    def __str__(self) -> str:
        if self.__reaction_name != "":
            return f"FAReaction: {self.__reaction_name}(C:{self.__C}, DB:{self.__DB}, OH:{self.__OH})"
        else:
            return f"FAReaction(C:{self.__C}, DB:{self.__DB}, OH:{self.__OH})"

    def __repr__(self) -> str:
        return self.__str__()



from __future__ import annotations
from typing import List
from .lipid import Lipid, FA
from .reference import ReferenceLipid
from .reaction import Reaction, FAReaction
from datetime import datetime


class Edge:

    __slots__ = ["__l1",
                 "__l2",
                 "__enzyme_id",
                 "__reaction_id",
                 "__reaction_type",
                 "__notes",
                 "__enzyme_uniprot",
                 "__enzyme_gene_name",
                 "__reaction_structure", # 1,1; 1,2; 2,2
                 "__l1_type",
                 "__l2_type",
                 "__set_id",
                 "__fa_param",
                 "__nl_participants"]

    def __init__(self, l1: Lipid, l2: Lipid, enzyme_id: str,
                 reaction_id: str, reaction_type: str, notes: str,
                 uniprot: str = "", gene_name: str = "",
                 reaction_structure: str = "1,1",
                 l1_type: str = "substrate",
                 l2_type: str = "product",
                 set_id: int = 0,
                 fa_param: str = "",
                 nl_participants: List[str] = None):
        if enzyme_id == None:
            raise ValueError("Enzyme ID cannot be None")
        self.__l1 = l1
        self.__l2 = l2
        self.__enzyme_id = enzyme_id
        self.__reaction_type = reaction_type
        self.__reaction_id = reaction_id
        self.__notes = notes
        self.__enzyme_gene_name = gene_name
        self.__enzyme_uniprot = uniprot
        self.__reaction_structure = reaction_structure
        self.__l1_type = l1_type
        self.__l2_type = l2_type
        self.__set_id = set_id
        self.__fa_param = fa_param
        if nl_participants is None:
            self.__nl_participants = []
        else:
            self.__nl_participants = nl_participants

    def get_l1(self) -> Lipid:
        return self.__l1

    def get_l2(self) -> Lipid:
        return self.__l2

    def get_enzyme_id(self, nlen=100, full=False) -> str:
        if full:
            return self.__enzyme_id
        elif len(self.__enzyme_id) > nlen:
            return self.__enzyme_id[0:nlen] + "..."
        else:
            return self.__enzyme_id

    def get_fa_param(self) -> str:
        return self.__fa_param

    def get_enzyme_repr(self, nlen=100) -> str:
        if self.__enzyme_gene_name == "" or self.__enzyme_gene_name is None:
            if len(self.get_enzyme_id()) > nlen:
                return self.get_enzyme_id()[0:nlen] + "..."
            else:
                return self.get_enzyme_id()
        else:
            return self.__enzyme_gene_name[0:nlen] + "..."

    def get_enzyme_uniprot(self) -> str:
        return self.__enzyme_uniprot

    def get_enzyme_gene_name(self) -> str:
        return self.__enzyme_gene_name

    def get_reaction_type(self) -> str:
        return self.__reaction_type

    def get_reaction_id(self) -> str:
        return self.__reaction_id

    def get_notes(self) -> str:
        return self.__notes

    def get_reaction_structure(self) -> str:
        return self.__reaction_structure

    def get_l1_type(self) -> str:
        return self.__l1_type

    def get_l2_type(self) -> str:
        return self.__l2_type

    def get_set_id(self) -> int:
        return self.__set_id

    def get_nl_participants(self) -> List[str]:
        return self.__nl_participants

    def __str__(self) -> str:

        # tmp_l = sorted([str(self.__l1), str(self.__l2)])
        return f"{str(self.__l1)} - {str(self.__l2)} [{self.__reaction_type}; {self.__reaction_id}, " \
               f"{self.get_enzyme_repr()}, {self.__notes}, {self.__reaction_structure}]"

    def sum_species_str(self) -> str:
        # tmp_l = sorted([self.__l1.sum_species_str(), self.__l2.sum_species_str()])
        return f"{str(self.__l1)} - {str(self.__l2)} [{self.__reaction_type}; {self.__reaction_id}, " \
               f"{self.get_enzyme_repr()}, {self.__notes}, {self.__reaction_structure}]"

    def native_str(self) -> str:
        tmp_l = [self.__l1.get_native_string(), self.__l2.get_native_string()]
        return f"{tmp_l[0]} - {tmp_l[1]} [{self.__reaction_type}; {self.__reaction_id}, " \
               f"{self.get_enzyme_repr()}, {self.__notes}, {self.__reaction_structure}]"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other: Edge) -> bool:
        if self.__hash__() == hash(other):
            return True
        else:
            return False

reaction_str = "Reaction: "
enzyme_str = "Enzyme: "

def edge_attr_dict(x: Edge):
    return {
        "enzyme_id": x.get_enzyme_id(full=True),
        "reaction_type": x.get_reaction_type(),
        "reaction_id": x.get_reaction_id(),
        "notes": x.get_notes(),
        "enzyme_gene_name": x.get_enzyme_gene_name(),
        "enzyme_raw_id": x.get_enzyme_repr(),
        "enzyme_uniprot": x.get_enzyme_uniprot(),
        "reaction_structure": x.get_reaction_structure(),
        "l1_type": x.get_l1_type(),
        "l2_type": x.get_l2_type(),
        "set_id": x.get_set_id(),
        "id_fa_raw": x.get_enzyme_id() + "\nFA: " + x.get_fa_param(),
        "id_fa_param": x.get_fa_param(),
        "nl_participants": x.get_nl_participants()
    }

def lipid_edges_to_bip(el: list, bipartite_type: str = "reaction"):
    new_l = []
    if bipartite_type == "reaction":
        for e in el:
            if e[2]["reaction_id"] == "":
                new_l.append(e)
            else:

                new_l.append([e[0], reaction_str + e[2]["reaction_id"], dict(e[2], enzyme_edge=True)])
                new_l.append([e[1], reaction_str + e[2]["reaction_id"], dict(e[2], enzyme_edge=True)])

        return new_l

    elif bipartite_type == "enzyme":
        for e in el:
            if e[2]["enzyme_id"] == "":
                new_l.append(e)
            else:
                new_l.append([e[0], enzyme_str + e[2]["enzyme_id"], dict(e[2], enzyme_edge=True)])
                new_l.append([e[1], enzyme_str + e[2]["enzyme_id"], dict(e[2], enzyme_edge=True)])

        return new_l

    elif bipartite_type == "hyper":
        for e in el:
            # Non hyper edges
            f = e[2].copy()
            if f["enzyme_id"] == "":
                f["hyper"] = False
                new_l.append([e[0], e[1], dict(f, enzyme_edge=False)])
            # Hyper edges
            else:
                f["hyper"] = True
                new_l.append([e[0], enzyme_str + f["enzyme_id"], dict(f, enzyme_edge=True)])
                new_l.append([e[1], enzyme_str + f["enzyme_id"], dict(f, enzyme_edge=True)])
                new_l.append([e[0], e[1], dict(f, enzyme_edge=False)])

        return new_l

    else:
        raise ValueError("'bipartite_type' must be one of ('reaction', 'enzyme', 'hyper')")

class Edgelist:

    __slots__ = ["__edges"]

    def __init__(self, edges: List[Edge] = None):
        if edges is None:
            self.__edges = []
        else:
            self.__edges = edges

    def raw_edgelist_copy(self) -> List[Edge]:
        return self.__edges.copy()

    def raw_enzymes(self) -> List[str]:
        return [x.get_enzyme_repr() for x in self.__edges]

    def raw_reaction_ids(self) -> List[str]:
        return [x.get_reaction_id() for x in self.__edges]

    def raw_edge_notes(self) -> List[str]:
        return [x.get_notes() for x in self.__edges]

    def raw_edge_types(self) -> List[str]:
        return [x.get_reaction_type() for x in self.__edges]

    def add_edge(self, edge: Edge):
        self.__edges.append(edge)

    def add_edges(self, edges: List[Edge]):
        self.__edges += edges

    def __add__(self, other: Edgelist):
        return Edgelist(edges=self.raw_edgelist_copy()+other.raw_edgelist_copy())

    def molecular_species_edgelist(self, filter_duplicates: bool = False,
                                   excluded_reaction_types: List[str] = None,
                                   bipartite: bool = False, bipartite_type: str = "reaction") -> List:
        """

        :param filter_duplicates:
        :param excluded_reaction_types:
        :param bipartite:
        :param bipartite_type: "reaction" or "enzyme"
        :return:
        """

        if not filter_duplicates:
            if excluded_reaction_types is None:
                return_l = [[str(x.get_l1()), str(x.get_l2()), edge_attr_dict(x)] for x in self.__edges]

            else:
                return_l = [[str(x.get_l1()), str(x.get_l2()), edge_attr_dict(x)] for x in self.__edges if
                        x.get_reaction_type() not in excluded_reaction_types]
        else:
            tmp_e = set(self.__edges)
            if excluded_reaction_types is None:
                return_l = [[str(x.get_l1()), str(x.get_l2()), edge_attr_dict(x)] for x in tmp_e]
            else:
                return_l = [[str(x.get_l1()), str(x.get_l2()), edge_attr_dict(x)] for x in tmp_e if
                        x.get_reaction_type() not in excluded_reaction_types]

        if not bipartite:
            return return_l
        else:
            return lipid_edges_to_bip(return_l, bipartite_type=bipartite_type)


    def __native_set(self) -> List[Edge]:
        unique_edges = []
        unique_edge_str = []
        for e in self.__edges:
            if e.native_str() in unique_edge_str:
                pass
            else:
                unique_edge_str.append(e.native_str())
                unique_edges.append(e)

        return unique_edges

    def native_edgelist(self, filter_duplicates: bool = False,
                        excluded_reaction_types: List[str] = None,
                                   bipartite: bool = False, bipartite_type: str = "reaction") -> List:
        start = datetime.now()
        if not filter_duplicates:
            if excluded_reaction_types is None:
                return_l = [[str(x.get_l1().get_native_string()), str(x.get_l2().get_native_string()),
                         edge_attr_dict(x)] for x in self.__edges]
            else:
                return_l = [[str(x.get_l1().get_native_string()), str(x.get_l2().get_native_string()),
                         edge_attr_dict(x)] for x in self.__edges if
                        x.get_reaction_type() not in excluded_reaction_types]
        else:
            tmp_e = self.__native_set()
            if excluded_reaction_types is None:
                return_l = [[str(x.get_l1().get_native_string()), str(x.get_l2().get_native_string()),
                         edge_attr_dict(x)] for x in tmp_e]
            else:
                return_l = [[str(x.get_l1().get_native_string()), str(x.get_l2().get_native_string()),
                         edge_attr_dict(x)] for x in tmp_e if
                        x.get_reaction_type() not in excluded_reaction_types]
        if not bipartite:
            return return_l
        else:
            tmp = lipid_edges_to_bip(return_l, bipartite_type=bipartite_type)
            return tmp

    def __sum_species_set(self) -> List[Edge]:
        unique_edges = []
        unique_edge_str = []
        for e in self.__edges:
            if e.sum_species_str() in unique_edge_str:
                pass
            else:
                unique_edge_str.append(e.sum_species_str())
                unique_edges.append(e)

        return unique_edges

    def sum_species_edgelist(self, filter_duplicates=False, excluded_reaction_types: List[str] = None,
                                   bipartite: bool = False, bipartite_type: str = "reaction") -> List:

        if not filter_duplicates:
            if excluded_reaction_types is None:
                return_l = [[x.get_l1().sum_species_str(), x.get_l2().sum_species_str(),
                         edge_attr_dict(x)] for x in self.__edges]
            else:
                return_l = [[x.get_l1().sum_species_str(), x.get_l2().sum_species_str(),
                         edge_attr_dict(x)] for x in self.__edges if
                        x.get_reaction_type() not in excluded_reaction_types]
        else:
            tmp_e = self.__sum_species_set()
            if excluded_reaction_types is None:
                return_l = [[x.get_l1().sum_species_str(), x.get_l2().sum_species_str(),
                         edge_attr_dict(x)] for x in tmp_e]
            else:
                return_l = [[x.get_l1().sum_species_str(), x.get_l2().sum_species_str(),
                         edge_attr_dict(x)] for x in tmp_e if
                        x.get_reaction_type() not in excluded_reaction_types]
        if not bipartite:
            return return_l
        else:
            return lipid_edges_to_bip(return_l, bipartite_type=bipartite_type)

    def __str__(self) -> str:
        return f"Edgelist with {len(self.__edges)} edges."

    def __len__(self) -> int:
        return len(self.__edges)
